from hadar.solver.input import *
from hadar.solver.output import Result


def assert_study(self, expected: Result, result: Result):
    for name, node in expected.nodes.items():
        if name not in result.nodes.keys():
            self.fail('Node {} expected but not'.format(name))
        res = result.nodes[name]

        # Consumptions
        for cons_expected, cons_res in zip(node.consumptions, res.consumptions):
            self.assertEqual(cons_expected.type, cons_res.type,
                             "Consumption for node {} has different type".format(name))
            np.testing.assert_array_equal(cons_expected.quantity, cons_res.quantity,
                             'Consumption {} for node {} has different quantity'.format(cons_expected.type, name))
            self.assertEqual(cons_expected.cost, cons_res.cost,
                             'Consumption {} for node {} has different cost'.format(cons_expected.type, name))

        # Productions
        for prod_expected, prod_res in zip(node.productions, res.productions):
            self.assertEqual(prod_expected.type, prod_res.type,
                             "Production for node {} has different type".format(name))
            np.testing.assert_array_equal(prod_expected.quantity, prod_res.quantity,
                             'Production {} for node {} has different quantity'.format(prod_expected.type, name))
            self.assertEqual(prod_expected.cost, prod_res.cost,
                             'Production {} for node {} has different cost'.format(prod_expected.type, name))

        # Borders
        for border_expected, border_res in zip(node.borders, res.borders):
            self.assertEqual(border_expected.dest, border_res.dest,
                             "Border for node {} has different type".format(name))
            np.testing.assert_array_equal(border_expected.quantity, border_res.quantity,
                             'Border {} for node {} has different quantity'.format(border_expected.dest, name))
            self.assertEqual(border_expected.cost, border_res.cost,
                             'Border {} for node {} has different cost'.format(border_expected.dest, name))


def plot(d):

    print('=============================================================================================')
    print("Node ", d.state.name, 'rac=', d.state.rac, 'cost=', d.state.cost)
    print('\nEvents')
    print('\ttype\tmes')
    for event in d.events:
        print('\t{type: <8}{mes}'.format(type=event.type, mes=event.message))

    print(d.state.consumptions)
    print(d.state.productions)
    print(d.state.exchanges)



import unittest
import pandas as pd

from hadar.solver.actor.domain.message import Exchange
from hadar.solver.actor.ledger import LedgerExchange, LedgerProduction, LedgerConsumption, LedgerBorder


class TestLedgerExchange(unittest.TestCase):

    def test(self):
        ex = [
            Exchange(id=1234, production_type='solar', quantity=10, path_node=['fr']),
            Exchange(id=9876, production_type='solar', quantity=10, path_node=['fr']),
            Exchange(id=5432, production_type='solar', quantity=10, path_node=['be']),
            Exchange(id=4566, production_type='nuclear', quantity=10, path_node=['fr'])
        ]
        ledger = LedgerExchange()
        ledger.add_all(ex, 'export')

        self.assertEqual(30, ledger.sum_production(production_type='solar'), "Wrong ledger behaviour")
        self.assertEqual(30, ledger.sum_border(name='fr'), 'Wrong ledger behaviour')

        ledger.delete(ex[1])
        self.assertEqual(20, ledger.sum_production(production_type='solar'), "Wrong ledger behaviour")
        self.assertEqual(20, ledger.sum_border(name='fr'), 'Wrong ledger behaviour')

    def test_filter_production_available(self):
        # Input
        ex = [
            Exchange(id=1, production_type='solar', quantity=1, path_node=['be']),
            Exchange(id=2, production_type='nuclear', quantity=1, path_node=['fr'])
        ]
        ledger = LedgerExchange()
        ledger.add_all(ex, 'export')

        prod = pd.DataFrame({'type': ['nuclear', 'wind'],
                             'quantity': [1, 1],
                             'used': [False, False],
                             'cost': [10, 10]},
                            index=[2, 3])

        # Expected
        prod_exp = pd.DataFrame({'type': ['wind'],
                                 'quantity': [1],
                                 'used': [False],
                                 'cost': [10]},
                                index=[3])

        # Test
        pd.testing.assert_frame_equal(prod_exp, ledger.filter_production_available(productions=prod))


class TestLedgerProduction(unittest.TestCase):

    def test(self):
        mock = MockUUID()
        ledger = LedgerProduction(uuid_generate=mock.generate)

        # Add production
        ledger.add_production(cost=10, quantity=2, type='nuclear')
        ledger.add_production(cost=10, quantity=3, type='solar', used=True)
        ex = Exchange(id=1234, production_type='wind', quantity=1, path_node=['fr'])
        ledger.add_exchanges(cost=10, ex=[ex])

        # Inspect ledger
        expectedA = pd.DataFrame({'cost': [10, 10, 10, 10, 10, 10],
                                  'quantity': [1, 1, 1, 1, 1, 1],
                                  'type': ['nuclear', 'nuclear', 'solar', 'solar', 'solar', 'import'],
                                  'used': [False, False, True, True, True, False],
                                  'path_node': [None, None, None, None, None, ['fr']]},
                                 index=[1, 2, 3, 4, 5, 1234])
        pd.testing.assert_frame_equal(expectedA, ledger.ledger)

        # Delete production
        ledger.delete(id=2)

        # Inspect ledger
        expectedB = pd.DataFrame({'cost': [10, 10, 10, 10, 10],
                                  'quantity': [1, 1, 1, 1, 1],
                                  'type': ['nuclear', 'solar', 'solar', 'solar', 'import'],
                                  'used': [False, True, True, True, False],
                                  'path_node': [None, None, None, None, ['fr']]},
                                 index=[1, 3, 4, 5, 1234])
        pd.testing.assert_frame_equal(expectedB, ledger.ledger)

    def test_group_free_productions(self):
        mock = MockUUID()
        ledger = LedgerProduction(uuid_generate=mock.generate)

        # Add production
        ledger.add_production(cost=10, quantity=2, type='nuclear')
        ledger.add_production(cost=20, quantity=10, type='wind')
        ledger.add_production(cost=10, quantity=3, type='solar', used=True)

        # Expected
        expected = pd.DataFrame(data={'cost': [10, 20],
                                      'quantity': [2, 10]},
                                index=['nuclear', 'wind'])
        expected.index.name = 'type'

        # Test
        pd.testing.assert_frame_equal(expected, ledger.group_free_productions())

    def test_get_production_quantity(self):
        mock = MockUUID()
        ledger = LedgerProduction(uuid_generate=mock.generate)

        # Add production
        ledger.add_production(cost=10, quantity=15, type='nuclear')
        ledger.add_production(cost=10, quantity=15, type='nuclear')
        ledger.add_production(cost=10, quantity=15, type='nuclear', used=True)
        ledger.add_production(cost=10, quantity=20, type='solar', used=True)

        self.assertEqual(30, ledger.get_production_quantity(type='nuclear', used=False))
        self.assertEqual(15, ledger.get_production_quantity(type='nuclear', used=True))
        self.assertEqual(20, ledger.get_production_quantity(type='solar', used=True))

    def test_apply_adequacy(self):
        # Input
        uuid_mock = MockUUID()
        ledger = LedgerProduction(uuid_generate=uuid_mock.generate)
        ledger.add_production(cost=10, quantity=5, type='solar')
        ledger.add_production(cost=15, quantity=3, type='nuclear')
        ledger.add_production(cost=5, quantity=2, type='wind')

        # Expected
        expected = pd.DataFrame({'cost': [5] * 2 + [10] * 5 + [15] * 3,
                                 'quantity': [1] * 10,
                                 'type': ['wind'] * 2 + ['solar'] * 5 + ['nuclear'] * 3,
                                 'used': [True] * 7 + [False] * 3,
                                 'path_node': [None] * 10},
                                index=[9, 10, 1, 2, 3, 4, 5, 6, 7, 8])

        # Test
        cost, power = ledger.apply_adequacy(quantity=7)

        self.assertEqual(60, cost)
        self.assertEqual(7, power)
        pd.testing.assert_frame_equal(expected, ledger.ledger)

    def test_split_quantity(self):
        # Input
        prod = pd.DataFrame({'quantity': [8, 6, 4],
                             'cost': [10, 10, 10],
                             'type': ['solar', 'solar', 'solar'],
                             'used': [False, False, False]},
                            index=[1, 2, 3])

        # Expected
        used = pd.DataFrame({'quantity': [8, 6],
                             'cost': [10, 10],
                             'type': ['solar', 'solar'],
                             'used': [False, False]},
                            index=[1, 2])

        free = pd.DataFrame({'quantity': [4],
                             'cost': [10],
                             'type': ['solar'],
                             'used': [False]},
                            index=[3])

        # Test
        res_used, res_free = LedgerProduction.split_by_quantity(prod=prod, quantity=15)
        pd.testing.assert_frame_equal(used, res_used)
        pd.testing.assert_frame_equal(free, res_free)


class TestLedgerConsumption(unittest.TestCase):
    def test(self):
        ledger = LedgerConsumption()

        # Add
        ledger.add(type='load', cost=20, quantity=100)
        ledger.add(type='stockage', cost=20, quantity=200)

        # Inspect
        expected = pd.DataFrame({'cost': [20, 20], 'quantity': [100, 200]}, index=['load', 'stockage'])
        pd.testing.assert_frame_equal(expected, ledger.ledger)

        # Delete
        ledger.delete(type='load')
        expected = pd.DataFrame({'cost': [20], 'quantity': [200]}, index=['stockage'])
        pd.testing.assert_frame_equal(expected, ledger.ledger)

    def test_compute_cost(self):
        ledger = LedgerConsumption()
        ledger.add(type='load', cost=10, quantity=100)

        self.assertEqual(0, ledger.compute_cost(quantity=150))
        self.assertEqual(500, ledger.compute_cost(quantity=50))

        ledger = LedgerConsumption()
        ledger.add(type='load1', cost=1, quantity=10)
        ledger.add(type='load2', cost=2, quantity=10)
        ledger.add(type='load3', cost=3, quantity=10)
        ledger.add(type='load4', cost=4, quantity=10)

        self.assertEqual(0, ledger.compute_cost(quantity=40))
        self.assertEqual(16, ledger.compute_cost(quantity=27))


class TestLedgerBorder(unittest.TestCase):
    def test(self):
        ledger = LedgerBorder()

        # Add
        ledger.add(dest='fr', cost=20, quantity=100)
        ledger.add(dest='be', cost=20, quantity=200)

        # Inspect
        expected = pd.DataFrame({'cost': [20, 20], 'quantity': [100, 200]}, index=['fr', 'be'])
        pd.testing.assert_frame_equal(expected, ledger.ledger)

        pd.testing.assert_series_equal(ledger.ledger.loc['fr'], ledger.find_border_in_path(['it', 'fr']))

        # Delete
        ledger.delete(dest='fr')
        expected = pd.DataFrame({'cost': [20], 'quantity': [200]}, index=['be'])
        pd.testing.assert_frame_equal(expected, ledger.ledger)


class MockUUID:
    def __init__(self):
        self.inc = 0

    def generate(self):
        self.inc += 1
        return self.inc



import uuid

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Tuple, List

from hadar.solver.actor.domain.message import *
from hadar.solver.actor.ledger import *
from hadar.solver.input import DTO


class State(DTO):
    """
    Represent current adequacy configuration. Each Handler has to update and forward this state.
    """
    def __init__(self, name: str, consumptions: LedgerConsumption, borders: LedgerBorder,
                 productions: LedgerProduction, rac: int, cost: int):
        """
        Create instance.

        :param name: node's name
        :param consumptions: consumptions list to match
        :param borders: borders capacity list
        :param productions: productions capacity list
        :param rac: remain capacity
        :param cost: global adequacy cost
        """
        self.name = name
        self.consumptions = consumptions
        self.borders = borders
        self.productions = productions
        self.exchanges = LedgerExchange()
        self.rac = rac
        self.cost = cost


class HandlerParameter:
    """Global parameters used by all handlers"""
    def __init__(self, ask=None, tell=None, uuid_generate: uuid = uuid.uuid4, min_exchange: int = 1):
        self.ask = ask
        self.tell = tell
        self.uuid_generate = uuid_generate


class Handler(ABC):
    """
    Represent an atomic behaviour. Handler update state object and call other handlers.
    Message receiving behaviour is implemented by chaining handler according to Chain of Responsabilities pattern
    """
    def __init__(self, params: HandlerParameter = None):
        self.params = params

    def set_params(self, params: HandlerParameter):
        self.params = params

    @abstractmethod
    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        pass


class ReturnHandler(Handler):
    """Stub handler chain. Return state value"""
    def __init__(self):
        Handler.__init__(self)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Return state and message without any change.

        :param state: current state
        :param message: message receive
        :return: same state, same message
        """
        return state, message


class CancelUselessImportationHandler(Handler):
    """
    Cancel exchanges find in not free productions
    """
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler.

        :param next: handler to execute at this end of process
        :param params: handler parameters (use only if handler is first of the chain)
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Execute process.

        :param state: current state
        :param message: message receive from dispatcher. (Not used by handler)
        :return: new state, response message. (Not response from handler)
        """
        useless = state.productions.filter_useless_exchanges()
        exs = state.exchanges.get_exchanges(useless.index.values)

        state.exchanges.delete_all(exs)
        state.productions.delete_all(useless.index)
        self._send_cancel_exchange(exs)

        return self.next.execute(deepcopy(state))

    def _send_cancel_exchange(self, exchanges: List[Exchange]):
        """
        Send canceled exchange order regrouped by producer.

        :param exchanges: exchanges to cancel
        :return:
        """
        productions = {}
        for ex in exchanges:
            path = tuple(ex.path_node)
            if path not in productions.keys():
                productions[path] = []
            productions[path].append(ex)

        for path, exchanges in productions.items():
            cancel = ConsumerCanceledExchange(exchanges=exchanges, path_node=list(path))
            self.params.tell(to=path[0], mes=cancel)


class ProposeFreeProductionHandler(Handler):
    """
    Send as proposal free production.
    Check already sent production quantity and border capacities.
    """
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler.

        :param next: handler to execute at this end of process
        :param params: handler parameters (use only if handler is first of the chain)
        """
        Handler.__init__(self, params)
        self.next = next
        self.forward = ForwardMessageHandler(next=ReturnHandler())
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)
        self.forward.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Execute process.

        :param state: current state
        :param message: message received from dispatcher. Not used by this handler
        :return: new state, response message (Not response from this handler)
        """
        for p_type, (p_cost, p_quantity) in state.productions.group_free_productions().iterrows():
            prod_sent = state.exchanges.sum_production(production_type=p_type)
            qt = p_quantity - prod_sent
            prop = Proposal(production_type=p_type, cost=p_cost, quantity=qt, path_node=[])

            self.forward.execute(deepcopy(state), deepcopy(prop))

        return self.next.execute(deepcopy(state))


class ForwardMessageHandler(Handler):
    """Forward message to borders according to remain capacity"""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler.

        :param next: handler to execute at this end of process
        :param params: handler parameters (use only if handler is first of the chain)
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Forward message to borders.

        :param state: current state
        :param message: message with quantity and path_node properties
        :return: same state, messages (not for this handler)
        """
        for b_id, (b_cost, b_quantity) in state.borders.ledger.iterrows():
            export = state.exchanges.sum_border(name=b_id)
            copy = deepcopy(message)
            copy.quantity = min(message.quantity, b_quantity - export)
            copy.cost += b_cost
            copy.path_node = [state.name] + message.path_node
            self.params.tell(to=b_id, mes=copy)

        return self.next.execute(deepcopy(state))


class CancelExportationHandler(Handler):
    """Cancel exchange."""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler.

        :param next: handler to execute after delete exchange.
        :param params: handler parameters (use only if handler is first of the chain)
        """
        Handler.__init__(self, params)
        self.next = next
        self.next.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Execute process.

        :param state: current state
        :param message: ConsumerCancelExchange message
        :return: new state, response message (Not response from handler)
        """
        # delete exchange in ledger
        state.exchanges.delete_all(message.exchanges)

        return self.next.execute(deepcopy(state), deepcopy(message))


class BackwardMessageHandler(Handler):
    """Forward message to border"""
    def __init__(self, after_backward: Handler, on_resume: Handler, type: str, params: HandlerParameter = None):
        """
        Create Handler
        :param after_backward: handler to execute after backward
        :param on_resume: handler to execute if not backward
        :param type: give what kind of connection 'ask' or 'tell'
        """
        Handler.__init__(self, params)
        self.type = type
        self.after_backward = after_backward
        self.on_resume = on_resume
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.after_backward.set_params(params)
        self.on_resume.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        if len(message.path_node) > 1:
            message.path_node = message.path_node[1:]
            if self.type == 'ask':
                res = self.params.ask(to=message.path_node[0], mes=message)
                return self.after_backward.execute(deepcopy(state), deepcopy(res))

            elif self.type == 'tell':
                self.params.tell(to=message.path_node[0], mes=message)
                return self.after_backward.execute(deepcopy(state), None)

        return self.on_resume.execute(deepcopy(state), deepcopy(message))


class AcceptExchangeHandler(Handler):
    """Create available exchanges according to proposal"""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create Handler.

        :param next: handler to call after process
        :param min_exchange: production are tokenized when exchanged. Set the tokenize size
        :param params: current handler parameters
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, offer: Any = None) -> Tuple[State, Any]:
        """
        Compute available exchanges according to production already sent.

        :param state: current state
        :param offer: ProposalOffer message
        :return: (state, [available exchanges])
        """
        prod_total = state.productions.get_free_productions_by_type(type=offer.production_type)
        prod_available = state.exchanges.filter_production_available(productions=prod_total)
        prod_needed, _ = LedgerProduction.split_by_quantity(prod_available, offer.quantity)
        ex = [Exchange(quantity=p_qt, id=p_id, production_type=p_type, path_node=offer.return_path_node)
              for p_id, (_, p_qt, p_type, _, _) in prod_needed.iterrows()]

        return self.next.execute(deepcopy(state), deepcopy(ex))


class SaveExchangeHandler(Handler):
    """Save exchange to ledger"""
    def __init__(self, next: Handler, exchange_type: str, params: HandlerParameter = None):
        """
        Create Handler

        :param next: handler to execute after save exchanges
        :param params: handler global paramaters
        :param exchange_type: which kind of exchanges is saved [import/export/transfer]
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)
        self.type = exchange_type

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Save exchanges in ledger.

        :param state: current state
        :param message: exchanges
        :return: new sate, same exchanges
        """

        for e in deepcopy(message):
            if e.quantity > 0:
                e.path_node = SaveExchangeHandler.trim_path(state.name, deepcopy(e.path_node))
                state.exchanges.add(e, self.type)
        return self.next.execute(deepcopy(state), deepcopy(message))
    
    @staticmethod
    def trim_path(name: str, path: List[str]):
        """
        trim uphill nodes in path.

        :param state: current state
        :param path: whole path from exchange producer
        :return: trimed path with only next nodes
        """
        try:
            i = path.index(name) + 1
        except ValueError:
            i = 0
        return path[i:]


class CheckOfferBorderCapacityHandler(Handler):
    """Check border capacity to respond to offer"""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler
        :param next: handler to execute after check border capacity
        :param params: parameters to use
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Check border capacity, update message with new quantity.

        :param state: current state
        :param message: ProposalOffer
        :return: (new state, response message)
        """
        border = state.borders.find_border_in_path(message.return_path_node)
        free_border_capacity = border.quantity - state.exchanges.sum_border(border.name)
        if not free_border_capacity:
            return deepcopy(state), []

        message.quantity = min(free_border_capacity, message.quantity)
        return self.next.execute(deepcopy(state), deepcopy(message))


class CompareNewProduction(Handler):
    """Compare cost with a new production"""
    def __init__(self, for_prod_useless: Handler, for_prod_useful: Handler, params: HandlerParameter = None):
        """
        Create Handler
        :param for_prod_useless: handler to execute if some of new production is useless
        :param for_prod_useful: handler to execute if some of new production is useful
        :param params: global parameters
        """
        Handler.__init__(self, params)
        self.for_prod_useless = for_prod_useless
        self.for_prod_useful = for_prod_useful
        self.adequacy = AdequacyHandler(next=ReturnHandler())
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.for_prod_useful.set_params(params)
        self.for_prod_useless.set_params(params)
        self.adequacy.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Compute state with new production, compare new cost.

        :param state: current state
        :param message: message received by dispatcher with quantity and cost attribute
        :return: state, message
        """
        test = deepcopy(state)
        test.productions.add_production(cost=message.cost, quantity=message.quantity, type='test')
        test, _ = self.adequacy.execute(test)
        if state.cost > test.cost:
            used_qt = test.productions.get_production_quantity(type='test', used=True)
            remain_qt = message.quantity - used_qt

            if remain_qt > 0:
                remain = deepcopy(message)
                remain.quantity = remain_qt
                self.for_prod_useless.execute(state=deepcopy(state), message=deepcopy(remain))

            message.quantity = used_qt

            return self.for_prod_useful.execute(deepcopy(state), deepcopy(message))
        else:
            return self.for_prod_useless.execute(deepcopy(state), deepcopy(message))


class MakerOfferHandler(Handler):
    """Respond to a proposal or forward proposal"""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler
        :param next: handler to execute when exchange from offer respond are received
        :param params: parameters to use
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, proposal: Any = None) -> Tuple[State, Any]:
        """
        Respond to proposal. Save new productions
        :param state: current state
        :param proposal: proposal to respond
        :return:
        """
        offer = ProposalOffer(production_type=proposal.production_type, cost=proposal.cost,
                              quantity=proposal.quantity,
                              path_node=proposal.path_node,
                              return_path_node=proposal.path_node[-2::-1] + [state.name])

        exchanges = self.params.ask(to=proposal.path_node[0], mes=offer)
        exchanges = [self.change_path(ex, proposal.path_node) for ex in exchanges]
        state.productions.add_exchanges(cost=proposal.cost, ex=exchanges)

        return self.next.execute(deepcopy(state), deepcopy(exchanges))

    def change_path(self, ex: Exchange, path_node: List[str]) -> Exchange:
        ex.path_node = path_node
        return ex


class AdequacyHandler(Handler):
    """Compute power mix to used for reach local adequacy"""
    def __init__(self, next: Handler, params: HandlerParameter = None):
        """
        Create handler
        :param next: handler to execute after check border capacity
        :param params: parameters to use
        """
        Handler.__init__(self, params)
        self.next = next
        self.set_params(params)

    def set_params(self, params: HandlerParameter):
        self.params = params
        self.next.set_params(params)

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Compute adequacy.

        :param state: current state
        :param message: message receive by dispatcher (not used by handler)
        :return: new state with new production stack & rac & cost, message send by dispatcher (not given by this handler)
        """
        load = state.consumptions.sum_quantity()
        cost, prod = state.productions.apply_adequacy(quantity=load)

        cost += state.consumptions.compute_cost(quantity=prod)

        state.cost = cost
        state.rac = prod - load

        return self.next.execute(state, message)



import logging
import time
import uuid
from typing import Callable, List

from pykka import ThreadingActor, ActorRegistry

from hadar.solver.output import *
from hadar.solver.input import *
from hadar.solver.actor.handler.entry import *
from hadar.solver.actor.handler.handler import AdequacyHandler, ReturnHandler, HandlerParameter, State
from hadar.solver.actor.ledger import *
from hadar.solver.actor.domain.message import *

class Waiter:
    """
    Waiter are updated by all actor at each message receive.
    Waiter determine when actor exchanges are finished, when anyone update it.
    """
    def __init__(self, wait_ms=0):
        """
        Create instance.

        :param wait_ms: waiting time for each loop
        """
        self.updated = True
        self.wait_ms = wait_ms

    def wait(self):
        """
        Endlessly loop. Stop only when actor don't notify waiter during a full time loop.

        :return:
        """
        while self.updated:
            self.updated = False
            time.sleep(self.wait_ms / 1000)

    def update(self):
        """
        Method used by actor to pock waiter that actors still exchange.

        :return:
        """
        self.updated = True


class Event:
    """
    Meta object to store message receive by actor.
    """
    def __init__(self, type: str, message, res=None):
        self.type = type
        self.message = message
        self.res = res


class Dispatcher(ThreadingActor):
    """
    Main actor. It optimize it local adequacy by exchange productions with its neighborhood
    """
    def __init__(self, name,
                 uuid_generate=uuid.uuid4,
                 waiter: Waiter = None,
                 input: InputNode = None):
        """
        Initiate actor.

        :param name: actor (node) name
        :param uuid_generate: uuid generate to use when tokenize production
        :param waiter: waiter instance to pock
        :param input: input node data to perform adequacy at best cost
        """
        super().__init__()

        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        # Save constructor params
        self.name = name
        self.consumptions = input.consumptions
        self.productions = input.productions
        self.borders = input.borders
        self.uuid_generate = uuid_generate

        # Instantiate output data
        self.out_node = OutputNode.build_like_input(input)

        # Set time horizon
        if self.consumptions:
            self.limit = self.consumptions[0].quantity.size
        elif self.productions:
            self.limit = self.productions[0].quantity.size
        elif self.borders:
            self.limit = self.borders[0].quantity.size
        self.t = 0

        # Setup waiter and and event sink
        self.waiter = waiter
        self.events = []

        # Setup handlers
        params = HandlerParameter(ask=self.ask_to, tell=self.tell_to, uuid_generate=self.uuid_generate)
        self.adequacy = AdequacyHandler(next=ReturnHandler())
        self.start = StartHandler(params=params)
        self.proposal = ProposalHandler(params=params)
        self.offer = ProposalOfferHandler(params=params)
        self.cancel_consumer_exchange = CanceledCustomerExchangeHandler(params=params)

        # Build first state
        self.state = Dispatcher.build_state(name=self.name, consumptions=self.consumptions, borders=self.borders,
                                            productions=self.productions, uuid_generate=self.uuid_generate, t=self.t)
        self.state, _ = self.adequacy.execute(self.state)

        # Registry actor
        self.actor_ref.actor_urn = name
        ActorRegistry.register(self.actor_ref)

    @staticmethod
    def build_state(name: str,
                    consumptions: List[Consumption],
                    borders: List[Border],
                    productions: List[Production],
                    uuid_generate: Callable, t: int) -> State:
        """
        Build new state according to timestamp.

        :param t: timestamp (i.e. quantity array index) to use
        :return: state with all ledgers setup with correct available quantities
        """
        # Build Consumers
        consumer_ledger = LedgerConsumption()
        for cons in consumptions:
            consumer_ledger.add(type=cons.type, cost=cons.cost, quantity=cons.quantity[t])

        # Build Producers
        producer_ledger = LedgerProduction(uuid_generate=uuid_generate)
        for prod in productions:
            producer_ledger.add_production(cost=prod.cost, quantity=prod.quantity[t], type=prod.type)

        # Build Border
        border_ledger = LedgerBorder()
        for b in borders:
            border_ledger.add(dest=b.dest, cost=b.cost, quantity=b.quantity[t])

        return State(name=name, consumptions=consumer_ledger, borders=border_ledger,
                     productions=producer_ledger, rac=0, cost=0)

    def on_receive(self, message):
        """
        Mail box of actor.

        :param message: next message to process
        :return:
        """
        self.logger.debug('%s receive message %s', self.name, message)

        res = None
        if isinstance(message, Start):
            self.state, _ = self.start.execute(self.state, message)

        elif isinstance(message, Snapshot):
            res = self

        elif isinstance(message, Next):
            res = self.name, self.next()

        elif isinstance(message, ProposalOffer):
            self.waiter.update()
            self.state, res = self.offer.execute(self.state, message)

        elif isinstance(message, Proposal):
            self.waiter.update()
            self.state, _ = self.proposal.execute(self.state, message)

        elif isinstance(message, ConsumerCanceledExchange):
            self.waiter.update()
            self.state, _ = self.cancel_consumer_exchange.execute(self.state, message)

        if res:
            self.logger.debug('%s response message %s', self.name, res)
        return res

    def on_stop(self):
        """
        Unregister actor
        :return:
        """
        ActorRegistry.unregister(self.actor_ref)

    def tell_to(self, to: str, mes):
        """
        Send message to asked actor using actor registry.

        :param to: actor name
        :param mes: message to send
        :return:
        """
        self.events.append(Event(type='tell', message=mes))
        ActorRegistry.get_by_urn(to).tell(mes)

    def ask_to(self, to: str, mes):
        """
        Send mess to asked actor using actor registry.

        :param to: actor name
        :param mes: message to send
        :return: actor response
        """
        self.events.append(Event(type='ask', message=mes))
        res = ActorRegistry.get_by_urn(to).ask(mes)
        self.events.append(Event(type='ask res', message=res))
        return res

    def next(self) -> OutputNode:
        """
        Handle next message. Update result. pass to next timestamp.

        :return: return OutputNode only if they have no more timestamp
        """
        self.out_node = self.compute_total(out_node=self.out_node, state=self.state, t=self.t)
        if self.t + 1 < self.limit:
            self.t += 1
            self.state = Dispatcher.build_state(name=self.name, consumptions=self.consumptions, borders=self.borders,
                                                productions=self.productions, uuid_generate=self.uuid_generate,
                                                t=self.t)
        else:
            return self.out_node

    @staticmethod
    def compute_total(out_node: OutputNode, state: State, t: int):
        """
        Compute dispatcher result:
        - copy initial consumption quantities
        - get production quantity used
        - sum production send to border

        :param out_node: OutputNode used by dispatcher to collect result
        :param state: current dispatcher state
        :param t: timestamp to update inside output result
        :return: new out_node with quantities updated
        """
        for i, _ in enumerate(out_node.consumptions):
            type = out_node.consumptions[i].type
            out_node.consumptions[i].quantity[t] = state.consumptions.find_consumption(type)['quantity']

        for i, _ in enumerate(out_node.productions):
            type = out_node.productions[i].type
            qt = state.productions.get_production_quantity(type=type, used=True)
            qt += state.exchanges.sum_production(production_type=type)
            out_node.productions[i].quantity[t] = qt

        for i, _ in enumerate(out_node.borders):
            dest = out_node.borders[i].dest
            out_node.borders[i].quantity[t] = state.exchanges.sum_border(name=dest)

        return out_node



import logging

from typing import List
from ortools.linear_solver.pywraplp import Solver

import hadar
from hadar.solver.lp.domain import *
from hadar.solver.input import InputNode, Study
from hadar.solver.lp.mapper import InputMapper, OutputMapper
from hadar.solver.output import OutputNode, Result

logger = logging.getLogger(__name__)


class ObjectiveBuilder:
    """
    Build objective cost function.
    """

    def __init__(self, solver: Solver):
        """
        Initiate new objective to minimize inside ortools solver.

        :param solver: ortools solver instance to use
        """
        self.objective = solver.Objective()
        self.objective.SetMinimization()
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def add_node(self, node: LPNode):
        """
        Add cost in objective for each node element.

        :param node: node to add
        :return:
        """
        self._add_consumption(node.consumptions)
        self._add_productions(node.productions)
        self._add_borders(node.borders)

    def _add_consumption(self, consumptions: List[LPConsumption]):
        """
        Add consumption cost. That mean we add cost of a loss of consumption.

        :param consumptions: consumption with loss variable and cost
        :return:
        """
        for cons in consumptions:
            self.objective.SetCoefficient(cons.variable, cons.cost)
            self.logger.debug('Add consumption %s into objective', cons.type)

    def _add_productions(self, prods: List[LPProduction]):
        """
        Add production cost. That mean the cost to use a production.

        :param prods: production with cost to use and used quantity variable
        :return:
        """
        for prod in prods:
            self.objective.SetCoefficient(prod.variable, prod.cost)
            self.logger.debug('Add production %s into objective', prod.type)

    def _add_borders(self, borders: List[LPBorder]):
        """
        Add border cost. That mean cost to use a border capacity.

        :param borders: borders with cost to use and used quantity variable
        :return:
        """
        for border in borders:
            self.objective.SetCoefficient(border.variable, border.cost)
            self.logger.debug('Add border %s->%s to objective', border.src, border.dest)

    def build(self):
        pass  # Currently nothing are need at the end. But we keep builder pattern syntax


class AdequacyBuilder:
    """
    Build adequacy flow constraint.
    """
    def __init__(self, solver: Solver, horizon: int):
        """
        Initiate.

        :param solver: ortools solver instance to use
        :param horizon: study horizon
        """
        self.constraints = [dict() for _ in range(horizon)]
        self.borders = [list() for _ in range(horizon)]
        self.solver = solver
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

    def add_node(self, name: str, t: int, node: LPNode):
        """
        Add flow constraint for a specific node.

        :param name: node name. Used to differentiate each equation
        :param t: timestamp index
        :param node: node to map constraint
        :return:
        """
        load = sum([c.quantity for c in node.consumptions])*1.0
        self.constraints[t][name] = self.solver.Constraint(load, load)

        self._add_consumptions(name, t, node.consumptions)
        self._add_productions(name, t, node.productions)
        self._add_borders(name, t, node.borders)

    def _add_consumptions(self, name: str, t: int, consumptions: List[LPConsumption]):
        """
        Add consumption flow. That mean loss of consumption is set a production to match
        equation in case there are not enough production.

        :param name: node's name
        :param t: timestamp index
        :param consumptions: consumptions with loss as variable
        :return:
        """
        for cons in consumptions:
            self.constraints[t][name].SetCoefficient(cons.variable, 1)
            self.logger.debug('Add lol %s for %s into adequacy constraint', cons.type, name)

    def _add_productions(self, name: str, t: int, productions: List[LPProduction]):
        """
        Add production flow. That mean production use is like a production.

        :param name: node's name
        :param t: timestamp index
        :param productions: productions with production used as variable
        :return:
        """
        for prod in productions:
            self.constraints[t][name].SetCoefficient(prod.variable, 1)
            self.logger.debug('Add prod %s for %s into adequacy constraint', prod.type, name)

    def _add_borders(self, name: str, t: int, borders: List[LPBorder]):
        """
        Add borders. That mean the border export is like a consumption.
        After all node added. The same export, become also an import for destination node.
        Therefore border has to be set like production for destination node.

        :param name: node's name
        :param t: timestamp index
        :param borders: border with export quantity as variable
        :return:
        """
        for bord in borders:
            self.borders[t].append(bord)
            self.constraints[t][name].SetCoefficient(bord.variable, -1)
            self.logger.debug('Add border %s for %s into adequacy constraint', bord.dest, name)

    def build(self):
        """
        Call when all node are added. Apply all import flow for each node.

        :return:
        """
        # Apply import border in adequacy
        for t in range(len(self.constraints)):
            for bord in self.borders[t]:
                self.constraints[t][bord.dest].SetCoefficient(bord.variable, 1)


def _solve(study: Study,
           solver: Solver,
           objective: ObjectiveBuilder,
           adequacy: AdequacyBuilder,
           in_mapper: InputMapper,
           out_mapper: OutputMapper) -> Result:
    """
    Solve adequacy flow problem with a linear optimizer.

    :param study: study to compute
    :param solver: solver to used
    :param objective: objective builder to use
    :param adequacy: adequacy builder to use
    :return: Result object with optimal solution
    """
    variables = [dict() for _ in range(study.horizon)]

    for t in range(0, study.horizon):
        for name, node in study.nodes.items():
            variables[t][name] = in_mapper.get_var(name=name, t=t)

            adequacy.add_node(name=name, t=t, node=variables[t][name])
            objective.add_node(node=variables[t][name])

    objective.build()
    adequacy .build()

    logger.info('Problem build. Start solver')
    solver.EnableOutput()
    solver.Solve()

    logger.info('Solver finish cost=%d', solver.Objective().Value())
    logger.debug(solver.ExportModelAsLpFormat(False).replace('\\', '').replace(',_', ','))

    for t in range(0, study.horizon):
        for name, node in study.nodes.items():
            out_mapper.set_var(name=name, t=t, vars=variables[t][name])

    return out_mapper.get_result()


def solve_lp(study: Study) -> Result:
    """
    Solve adequacy flow problem with a linear optimizer.

    :param study: study to compute
    :return: Result object with optimal solution
    """
    solver = Solver('simple_lp_program', Solver.GLOP_LINEAR_PROGRAMMING)

    objective = ObjectiveBuilder(solver=solver)
    adequacy = AdequacyBuilder(solver=solver, horizon=study.horizon)

    in_mapper = InputMapper(solver=solver, study=study)
    out_mapper = OutputMapper(solver=solver, study=study)

    return _solve(study, solver, objective, adequacy, in_mapper, out_mapper)



from typing import Dict, List

import plotly.graph_objects as go
from matplotlib.cm import coolwarm

from hadar.aggregator.result import *
from hadar.viewer.abc import ABCPlotting


class HTMLPlotting(ABCPlotting):
    """
    Plotting implementation interactive html graphics. (Use plotly)
    """
    def __init__(self, agg, unit_symbol: str = '',
                 time_start=None, time_end=None,
                 cmap=coolwarm,
                 node_coord: Dict[str, List[float]] = None,
                 map_element_size: int = 1):
        """
        Create instance.

        :param agg: ResultAggragator instence to use
        :param unit_symbol: symbol on quantity unit used. ex. MW, litter, Go, ...
        :param time_start: time to use as the start of study horizon
        :param time_end: time to use as the end of study horizon
        :param cmap: matplotlib color map to use (coolwarm as default)
        :param node_coord: nodes coordinates to use for map plotting
        :param map_element_size: size on element draw on map. default as 1.
        """

        self.agg = agg
        self.unit = '(%s)' % unit_symbol if unit_symbol != '' else ''
        self.coord = node_coord
        self.size = map_element_size

        # Create time_index
        time = [time_start is None, time_end is None]
        if time == [True, False] or time == [False, True]:
            raise ValueError('You have to give both time_start and time_end')
        elif time == [False, False]:
            self.time_index = pd.date_range(start=time_start, end=time_end, periods=self.agg.horizon)
        else:
            self.time_index = np.arange(self.agg.horizon)

        # Create colors scale
        self.cmap = cmap
        self.cmap_plotly = HTMLPlotting.matplotlib_to_plotly(cmap, 255)

        self.cmap_cons = ['brown', 'blue', 'darkgoldenrod', 'darkmagenta', 'darkorange', 'cadetblue', 'forestgreen',
                          'indigo', 'olive', 'darkred']

    @classmethod
    def matplotlib_to_plotly(cls, cmap, res: int):
        """
        Convert matplotlib color scale to plotly color scale.

        :param cmap: matplotlib color scale function
        :param res: resolution to use
        :return: list of string use by plotly
        """
        h = 1.0 / (res - 1)
        pl_colorscale = []
        for k in range(res):
            C = (np.array(cmap(k * h)[:3]) * 255).astype(np.uint8)
            pl_colorscale.append([k * h, 'rgb' + str((C[0], C[1], C[2]))])
        return pl_colorscale

    def stack(self, node: str, prod_kind: str = 'used', cons_kind: str = 'asked'):
        """
        Plot with production stacked with area and consumptions stacked by dashed lines.

        :param node: select node to plot. If None, use a dropdown menu to select inside notebook
        :param prod_kind: select which prod to stack : available ('avail') or 'used'
        :param cons_kind: select which cons to stacl : 'asked' or 'given'
        :return: plotly figure or jupyter widget to plot
        """
        fig = go.Figure()
        c, p, b = self.agg.get_elements_inside(node=node)
        stack = np.zeros(self.agg.horizon)

        # stack production with area
        if p > 0:
            prod = self.agg.agg_prod(NodeIndex(node), TypeIndex(), TimeIndex()).sort_values('cost', ascending=True)
            for i, type in enumerate(prod.index.get_level_values('type').unique()):
                stack += prod.loc[type][prod_kind].sort_index().values
                fig.add_trace(go.Scatter(x=self.time_index, y=stack.copy(), name=type, mode='none',
                                         fill='tozeroy' if i == 0 else 'tonexty'))

        # add import in production stack
        balance = self.agg.get_balance(node=node)
        im = -np.clip(balance, None, 0)
        if not (im == 0).all():
            stack += im
            fig.add_trace(go.Scatter(x=self.time_index, y=stack.copy(), name='import', mode='none', fill='tonexty'))

        # Reset stack
        stack = np.zeros_like(stack)
        cons_lines = []
        # Stack consumptions with line
        if c > 0:
            cons = self.agg.agg_cons(NodeIndex(node), TypeIndex(), TimeIndex()).sort_values('cost', ascending=False)
            for i, type in enumerate(cons.index.get_level_values('type').unique()):
                stack += cons.loc[type][cons_kind].sort_index().values
                cons_lines.append([type, stack.copy()])

        # Add export in consumption stack
        exp = np.clip(balance, 0, None)
        if not (exp == 0).all():
            stack += exp
            cons_lines.append(['export', stack.copy()])

        # Plot line in the reverse sens to avoid misunderstood during graphics analyze
        for i, (type, stack) in enumerate(cons_lines[::-1]):
            fig.add_trace(go.Scatter(x=self.time_index, y=stack.copy(), line_color=self.cmap_cons[i % 10],
                                     name= type, line=dict(width=2)))

        fig.update_layout(title_text='Stack for node %s' % node,
                          yaxis_title="Quantity %s" % self.unit, xaxis_title="time")
        return fig

    def _plot_links(self, fig: go.Figure, start: str, end: str, color: str, qt: float):
        """
        Plot line with arrow to a figure.

        :param fig: figure to use
        :param start: start node
        :param end: end node
        :param color: color to use
        :param qt: quantity to set inside label
        :return:
        """
        S = np.array([self.coord[start][0], self.coord[start][1]])
        E = np.array([self.coord[end][0], self.coord[end][1]])

        # plot line
        fig.add_trace(go.Scattergeo(lat=[S[1], E[1]], hoverinfo='skip',
                                    lon=[S[0], E[0]], mode='lines',
                                    line=dict(width=4 * self.size, color=color)))
        # vector flow direction
        v = E - S
        n = np.linalg.norm(v)
        # Get orthogonal vector
        w = np.array([v[1], -v[0]])
        # Compute triangle points
        A = E - v * 0.5
        B = A - v / n * self.size * 0.5 - w / n * self.size * 0.25
        C = A - v / n * self.size * 0.5 + w / n * self.size * 0.25

        # plot arrow
        fig.add_trace(go.Scattergeo(lat=[B[1], A[1], C[1]], hoverinfo='text',
                                    lon=[B[0], A[0], C[0]], text=str(qt), mode='lines',
                                    line=dict(width=4 * self.size, color=color)))

    def exchanges_map(self, t: int, limit: int = None):
        """
        Plot a map with node (color are balance) and arrow between nodes (color for quantity).

        :param t: timestep to plot
        :return: plotly figure or jupyter widget to plot
        """
        if self.coord is None:
            raise ValueError('Please provide node coordinate by setting param node_coord in Plotting constructor')

        balances = [self.agg.get_balance(node=node)[t] for node in self.agg.nodes]
        if limit is None:
            limit = max(max(balances), -min(balances))

        fig = go.Figure()

        # plot links
        borders = self.agg.agg_border(SrcIndex(), DestIndex(), TimeIndex())
        for src in borders.index.get_level_values('src').unique():
            for dest in borders.loc[src].index.get_level_values('dest').unique():
                exchange = borders.loc[src, dest, t]['used']  # forward
                exchange -= borders.loc[dest, src, t]['used'] if (dest, src, t) in borders.index else 0  # backward

                color = 'rgb' + str(self.cmap(abs(exchange) / 2 / limit + 0.5)[:-1])
                if exchange > 0:
                    self._plot_links(fig=fig, start=src, end=dest, color=color, qt=exchange)
                else:
                    self._plot_links(fig=fig, start=dest, end=src, color=color, qt=-exchange)

        # plot nodes
        text = ['%s: %i' % (n, b) for n, b in zip(self.agg.nodes, balances)]
        lon = [self.coord[node][0] for node in self.agg.nodes]
        lat = [self.coord[node][1] for node in self.agg.nodes]

        fig.add_trace(go.Scattergeo(lon=lon, lat=lat, hoverinfo='text', text=text, mode='markers',
                                    marker=dict(size=15 * self.size,
                                                colorscale=self.cmap_plotly, cmin=-limit, color=balances,
                                                cmax=limit, colorbar_title="Net Position %s" % self.unit)))
        # Config plot
        fig.update_layout(title_text='Exchanges Map', showlegend=False, height=600,
                          geo=dict(projection_type='equirectangular', showland=True, showcountries=True,
                                   resolution=50, landcolor='rgb(200, 200, 200)', countrycolor='rgb(0, 0, 0)',
                                   fitbounds='locations'))

        return fig


import unittest

from hadar.solver.actor.actor import Dispatcher, State
from hadar.solver.actor.ledger import *
from hadar.solver.output import *


class TestDispatcher(unittest.TestCase):

    def test_build_state(self):
        # Input

        # Expected
        cons = LedgerConsumption()
        cons.add(type='load', cost=20, quantity=10)

        prod = LedgerProduction(uuid_generate=lambda: 42)
        prod.add_production(type='solar', cost=23, quantity=10)

        border = LedgerBorder()
        border.add(dest='be', cost=2, quantity=10)

        exp_state = State(name='fr', consumptions=cons, productions=prod, borders=border, rac=0, cost=0)

        # Test
        res_state = Dispatcher.build_state(name='fr',
                                      uuid_generate=lambda: 42,
                                      consumptions=[Consumption(type='load', quantity=np.array([10]), cost=20)],
                                      productions=[Production(type='solar', quantity=np.array([10]), cost=23)],
                                      borders=[Border(dest='be', quantity=np.array([10]), cost=2)], t=0)

        self.assertEqual(exp_state, res_state)

    def test_compute_total(self):
        mock_id = MockUUID()

        # Input
        ledger_cons = LedgerConsumption()
        ledger_cons.add(type='load', cost=2, quantity=7)

        ledger_prod = LedgerProduction()
        ledger_prod.add_production(type='solar', cost=3, quantity=5, used=True)
        ledger_prod.add_production(type='nuclear', cost=3, quantity=2, used=True)
        ledger_prod.add_production(type='nuclear', cost=3, quantity=3, used=False)

        ledger_border = LedgerBorder()
        ledger_border.add(dest='be', cost=2, quantity=7)

        state = State(name='fr', consumptions=ledger_cons, borders=ledger_border,
                                 productions=ledger_prod, cost=0, rac=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add(Exchange(id=0, production_type='nuclear', quantity=3, path_node=['be']),
                                       type='export')

        out_node = OutputNode(consumptions=[], borders=[], productions=[])
        out_node.consumptions = [OutputConsumption(type='load', quantity=np.array([0, 0]), cost=2)]
        out_node.productions = [OutputProduction(type='solar', cost=3, quantity=np.array([0, 0])),
                                OutputProduction(type='nuclear', cost=3, quantity=np.array([0, 0]))]
        out_node.borders = [OutputBorder(dest='be', cost=2, quantity=np.array([0, 0]))]

        # Expected
        expected = OutputNode(consumptions=[], borders=[], productions=[])
        expected.consumptions = [OutputConsumption(type='load', quantity=np.array([7, 0]), cost=2)]
        expected.productions = [OutputProduction(type='solar', cost=3, quantity=np.array([5, 0])),
                                OutputProduction(type='nuclear', cost=3, quantity=np.array([5, 0]))]
        expected.borders = [OutputBorder(dest='be', cost=2, quantity=np.array([3, 0]))]

        # Start test
        res = Dispatcher.compute_total(out_node=out_node, state=state, t=0)

        self.assertEqual(expected, res)


class MockUUID:
    def __init__(self):
        self.inc = 0

    def generate(self):
        self.inc += 1
        return self.inc



from typing import Union, TypeVar, List, Generic, Tuple, Type

import pandas as pd
import numpy as np

from hadar.solver.output import Result, Study

T = TypeVar('T')


class Index(Generic[T]):
    """
    Generic Index to use to select and rank data.
    """
    def __init__(self, column, index: Union[List[T], T] = None):
        """
        Initiate instance.

        :param column: column name link to this index
        :param index: list of index or element to filter from data. None by default to say keep all data.
        """
        self.all = False
        self.column = column
        if index is None:
            self.all = True
        elif not isinstance(index, list):
            self.index = [index]
        else:
            self.index = index

    def filter(self, df: pd.DataFrame) -> pd.Series:
        """
        Filter dataframe. Filter columns with columns attributes with index values.

        :param df: dataframe to filter
        :return: Series of boolean to set row to keep
        """
        if self.all:
            return df[self.column].notnull()
        return df[self.column].isin(self.index)

    def is_alone(self):
        """
        Ask if index filter element is alone.

        :return: if index filter only one value return True else False
        """
        return not self.all and len(self.index) <= 1


class NodeIndex(Index[str]):
    """Index implementation to filter nodes"""
    def __init__(self, index: Union[List[str], str] = None):
        Index.__init__(self, column='node', index=index)


class SrcIndex(Index[str]):
    """Index implementation to filter src node"""
    def __init__(self, index: Union[List[str], str] = None):
        Index.__init__(self, column='src', index=index)


class DestIndex(Index[str]):
    """Index implementation to filter destination node"""
    def __init__(self, index: Union[List[str], str] = None):
        Index.__init__(self, column='dest', index=index)


class TypeIndex(Index[str]):
    """Index implementation to filter type of elements"""
    def __init__(self, index: Union[List[str], str] = None):
        Index.__init__(self, column='type', index=index)


class TimeIndex(Index[int]):
    """Index implementation to filter by time step"""
    def __init__(self, index: Union[List[int], int] = None, start: int = None, end: int = None):
        """
        Create instance.

        :param index: one element or list on element to filter.
        :param start: start datetime to filter (to use instead of index)
        :param end: end datetime to filter (to use instead of index)
        """
        if index is None and start is None and end is None:
            Index.__init__(self, column='t')
        elif index is None:
            if start is None:
                raise ValueError('Please give an start index')
            if end is None:
                raise ValueError('Please give an end index')
            index = list(range(start, end))
        Index.__init__(self, column='t', index=index)


class ResultAggregator:
    """
    Single object to encapsulate all postprocessing aggregation.
    """
    def __init__(self, study: Study, result: Result):
        """
        Create an instance.

        :param study: study to use
        :param result: result of study used
        """
        self.result = result
        self.study = study

        self.consumption = self._build_consumption()
        self.production = self._build_production()
        self.border = self._build_border()

    def _build_consumption(self):
        """
        Flat all data to build global consumption dataframe
        columns: | cost | type | node | asked | given | t |
        """
        h = self.study.horizon
        s = h * sum([len(n.consumptions) for n in self.result.nodes.values()])
        cons = {'cost': np.empty(s), 'asked': np.empty(s), 'given': np.empty(s),
                'type': np.empty(s), 'node': np.empty(s), 't': np.empty(s)}
        cons = pd.DataFrame(data=cons)

        n_cons = 0
        for n, name in enumerate(self.result.nodes.keys()):
            for i, c in enumerate(self.result.nodes[name].consumptions):
                slices = cons.index[n_cons * h: (n_cons + 1) * h]
                cons.loc[slices, 'cost'] = c.cost
                cons.loc[slices, 'type'] = c.type
                cons.loc[slices, 'node'] = name
                cons.loc[slices, 'asked'] = self.study.nodes[name].consumptions[i].quantity
                cons.loc[slices, 'given'] = c.quantity
                cons.loc[slices, 't'] = np.arange(h)

                n_cons += 1

        return cons

    def _build_production(self):
        """
        Flat all data to build global production dataframe
        columns: | cost | avail | used | type | node | t |
        """
        h = self.study.horizon
        s = h * sum([len(n.productions) for n in self.result.nodes.values()])
        prod = {'cost': np.empty(s), 'avail': np.empty(s), 'used': np.empty(s),
                'type': np.empty(s), 'node': np.empty(s), 't': np.empty(s)}
        prod = pd.DataFrame(data=prod)

        n_prod = 0
        for n, name in enumerate(self.result.nodes.keys()):
            for i, c in enumerate(self.result.nodes[name].productions):
                slices = prod.index[n_prod * h: (n_prod + 1) * h]
                prod.loc[slices, 'cost'] = c.cost
                prod.loc[slices, 'type'] = c.type
                prod.loc[slices, 'node'] = name
                prod.loc[slices, 'avail'] = self.study.nodes[name].productions[i].quantity
                prod.loc[slices, 'used'] = c.quantity
                prod.loc[slices, 't'] = np.arange(h)

                n_prod += 1

        return prod

    def _build_border(self):
        """
        Flat all data to build global border dataframe
        columns: | cost | avail | used | src | dest | t |
        """
        h = self.study.horizon
        s = h * sum([len(n.borders) for n in self.result.nodes.values()])
        border = {'cost': np.empty(s), 'avail': np.empty(s), 'used': np.empty(s),
                  'src': np.empty(s), 'dest': np.empty(s), 't': np.empty(s)}
        border = pd.DataFrame(data=border)

        n_border = 0
        for n, name in enumerate(self.result.nodes.keys()):
            for i, c in enumerate(self.result.nodes[name].borders):
                slices = border.index[n_border * h: (n_border + 1) * h]
                border.loc[slices, 'cost'] = c.cost
                border.loc[slices, 'dest'] = c.dest
                border.loc[slices, 'src'] = name
                border.loc[slices, 'avail'] = self.study.nodes[name].borders[i].quantity
                border.loc[slices, 'used'] = c.quantity
                border.loc[slices, 't'] = np.arange(h)

                n_border += 1

        return border

    @staticmethod
    def _remove_useless_index_level(df: pd.DataFrame, indexes: List[Index]) -> pd.DataFrame:
        """
        If top index level has only on element then remove this index level. Do it recursively.

        :param df: dataframe with multi-index
        :param indexes: indexes level used
        :return: dataframe trimed with useless index(es) level
        """
        # TODO refactor without the help of indexes
        if indexes[0].is_alone() and indexes[0].index[0] in df.index:
            df = df.loc[indexes[0].index[0]].copy()
            return ResultAggregator._remove_useless_index_level(df, indexes[1:])
        else:
            return df

    @staticmethod
    def _pivot(i0: Index, i1: Index, i2: Index, df: pd.DataFrame) -> pd.DataFrame:
        """
        Pivot table by appling filter and index hirarchy asked by indexes.

        :param i0: first level index
        :param i1: second level index
        :param i2: third level index
        :param df: dataframe to pivot
        :return: pivot table
        """
        indexes = [i0.column, i1.column, i2.column]
        pt = pd.pivot_table(data=df[i0.filter(df) & i1.filter(df) & i2.filter(df)],
                            index=indexes, aggfunc=lambda x: x.iloc[0])

        return ResultAggregator._remove_useless_index_level(df=pt, indexes=[i0, i1, i2])

    @staticmethod
    def _assert_index(i0: Index, i1: Index, i2: Index, type: Type):
        """
        Check indexes cohesion. Raise ValueError exception if indexes are wrong.

        :param i0: first level index
        :param i1: second level index
        :param i2: third level index
        :param type: type to check inside index
        :return:
        """
        if not (isinstance(i0, type) or isinstance(i1, type) or isinstance(i2, type)):
            raise ValueError('Indexes must contain a {}'.format(type.__class__.__name__))

    def agg_cons(self, i0: Index, i1: Index, i2: Index) -> pd.DataFrame:
        """
        Aggregate consumption according to index level and filter.

        :param i0: first level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :param i1: second level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :param i2: third level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :return: dataframe with hierarchical and filter index level asked
        """
        ResultAggregator._assert_index(i0, i1, i2, TimeIndex)
        ResultAggregator._assert_index(i0, i1, i2, NodeIndex)
        ResultAggregator._assert_index(i0, i1, i2, TypeIndex)

        return ResultAggregator._pivot(i0, i1, i2, self.consumption)

    def agg_prod(self, i0: Index, i1: Index, i2: Index) -> pd.DataFrame:
        """
        Aggregate production according to index level and filter.

        :param i0: first level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :param i1: second level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :param i2: third level index. Index type must be [NodeIndex, TypeIndex, TimeIndex]
        :return: dataframe with hierarchical and filter index level asked
        """
        ResultAggregator._assert_index(i0, i1, i2, TimeIndex)
        ResultAggregator._assert_index(i0, i1, i2, NodeIndex)
        ResultAggregator._assert_index(i0, i1, i2, TypeIndex)

        return ResultAggregator._pivot(i0, i1, i2, self.production)

    def agg_border(self, i0: Index, i1: Index, i2: Index) -> pd.DataFrame:
        """
        Aggregate border according to index level and filter.

        :param i0: first level index. Index type must be [DestIndex, SrcIndex, TimeIndex]
        :param i1: second level index. Index type must be [DestIndex, SrcIndex, TimeIndex]
        :param i2: third level index. Index type must be [DestIndex, SrcIndex, TimeIndex]
        :return: dataframe with hierarchical and filter index level asked
        """
        ResultAggregator._assert_index(i0, i1, i2, TimeIndex)
        ResultAggregator._assert_index(i0, i1, i2, SrcIndex)
        ResultAggregator._assert_index(i0, i1, i2, DestIndex)

        return ResultAggregator._pivot(i0, i1, i2, self.border)

    def get_elements_inside(self, node: str):
        """
        Get numbers of elements by node.

        :param node: node name
        :return: (nb of consumptions, nb of productions, nb of border (export))
        """
        return len(self.result.nodes[node].consumptions),\
               len(self.result.nodes[node].productions),\
               len(self.result.nodes[node].borders)

    def get_balance(self, node: str) -> np.ndarray:
        """
        Compute balance over time on asked node.

        :param node: node asked
        :return: timeline array with balance exchanges value
        """
        balance = np.zeros(self.study.horizon)

        im = pd.pivot_table(self.border[self.border['dest'] == node][['used', 't']], index='t', aggfunc=np.sum)
        if im.size > 0:
            balance += -im['used'].values

        exp = pd.pivot_table(self.border[self.border['src'] == node][['used', 't']], index='t', aggfunc=np.sum)
        if exp.size > 0:
            balance += exp['used'].values
        return balance

    def get_cost(self, node: str) -> np.ndarray:
        cost = np.zeros(self.horizon)
        c, p, b = self.get_elements_inside(node)
        if c:
            cons = self.agg_cons(NodeIndex(node), TimeIndex(), TypeIndex())
            cost += ((cons['asked'] - cons['given'])*cons['cost']).groupby(axis=0, level=0).sum().sort_index().values

        if p:
            prod = self.agg_prod(NodeIndex(node), TimeIndex(), TypeIndex())
            cost += (prod['used']*prod['cost']).groupby(axis=0, level=0).sum().sort_index().values

        if b:
            border = self.agg_border(SrcIndex(node), TimeIndex(), DestIndex())
            cost += (border['used']*border['cost']).groupby(axis=0, level=0).sum().sort_index().values

        return cost

    @property
    def horizon(self) -> int:
        """
        Shortcut to get study horizon.

        :return: study horizon
        """
        return self.study.horizon

    @property
    def nodes(self) -> List[str]:
        """
        Shortcut to get list of node names

        :return: nodes name
        """
        return self.result.nodes.keys()



from typing import Dict, List

import matplotlib
import ipywidgets as widgets
from IPython.display import display, clear_output

from hadar.aggregator.result import *
from hadar.viewer.html import HTMLPlotting


class JupyterPlotting(HTMLPlotting):
    """
    Plotting implementation to use with Jupyter.
    Graphics are generated by HTMLPlotting, then jupyter widgets are used to be more flexible.
    """
    def __init__(self, agg, unit_symbol: str = '',
                 time_start=None, time_end=None,
                 cmap=matplotlib.cm.coolwarm,
                 node_coord: Dict[str, List[float]] = None,
                 map_element_size: int = 1):
        """
        Create instance.

        :param agg: ResultAggragator instance to use
        :param unit_symbol: symbol on quantity unit used. ex. MW, litter, Go, ...
        :param time_start: time to use as the start of study horizon
        :param time_end: time to use as the end of study horizon
        :param cmap: matplotlib color map to use (coolwarm as default)
        :param node_coord: nodes coordinates to use for map plotting
        :param map_element_size: size on element draw on map. default as 1.
        """

        HTMLPlotting.__init__(self, agg, unit_symbol, time_start, time_end, cmap, node_coord, map_element_size)

    def _dropmenu(self, plot, items, **kargs):
        """
        Wrap html graphics with dropdown menu.

        :param plot: plot function to call when value change
        :param items: list of items present in drop down menu
        :return:
        """
        menu = widgets.Dropdown(options=items, value=items[0],
                                description='Node:', disabled=False)
        output = widgets.Output()

        def _plot(select):
            with output:
                clear_output()
                fig = plot(self, select, **kargs)
                fig.show()

        def _on_event(event):
            if event['name'] == 'value' and event['type'] == 'change':
                _plot(event['new'])

        menu.observe(_on_event)
        display(menu, output)
        _plot(items[0])

    def stack(self, node: str = None, prod_kind: str = 'used', cons_kind: str = 'asked'):
        """
        Plot with production stacked with area and consumptions stacked by dashed lines.

        :param node: select node to plot. If None, use a dropdown menu to select inside notebook
        :param prod_kind: select which prod to stack : available ('avail') or 'used'
        :param cons_kind: select which cons to stacl : 'asked' or 'given'
        :return: plotly figure or jupyter widget to plot
        """
        if node is not None:
            return HTMLPlotting.stack(self, node, prod_kind, cons_kind).show()
        else:
            nodes = list(self.agg.nodes)
            self._dropmenu(HTMLPlotting.stack, nodes, prod_kind=prod_kind, cons_kind=cons_kind)

    def _intslider(self, plot, size):
        """
        Wrap plot with a intslider.

        :param plot: plot to call when value change
        :param size: size of intslider (min=0, step=1)
        :return:
        """
        slider = widgets.IntSlider(value=0, min=0, max=size, step=1, description='Timestep:', disabled=False,
                                   continuous_update=False, orientation='horizontal', readout=True, readout_format='d')
        output = widgets.Output()

        def _plot(select):
            with output:
                clear_output()
                fig = plot(self, select)
                fig.show()

        def _on_event(event):
            if event['name'] == 'value' and event['type'] == 'change':
                _plot(event['new'])

        slider.observe(_on_event)
        display(slider, output)
        _plot(0)

    def exchanges_map(self, t: int = None):
        """
        Plot a map with node (color are balance) and arrow between nodes (color for quantity).

        :param t: timestep to plot
        :return: plotly figure or jupyter widget to plot
        """
        if t is not None:
            return HTMLPlotting.exchanges_map(self, t).show()
        else:
            h = self.agg.horizon -1
            self._intslider(HTMLPlotting.exchanges_map, h)



from unittest.mock import MagicMock, call

import pandas as pd
import numpy as np
import unittest

from hadar.solver.actor.actor import Exchange
from hadar.solver.actor.handler.handler import State
from hadar.solver.actor.handler.handler import *
from hadar.solver.actor.ledger import *


class TestCancelExchangeUselessHandler(unittest.TestCase):

    def test_execute(self):
        # Create mock
        tell_mock = MagicMock()

        # Input
        exchange0 = Exchange(id=0, production_type='solar', quantity=1, path_node=['fr'])
        exchange1 = Exchange(id=1, production_type='solar', quantity=1, path_node=['fr'])
        exchange2 = Exchange(id=2, production_type='solar', quantity=1, path_node=['fr'])
        exchange3 = Exchange(id=3, production_type='solar', quantity=1, path_node=['be'])
        exchange4 = Exchange(id=4, production_type='solar', quantity=1, path_node=['be'])

        productions = LedgerProduction(uuid_generate=lambda: None)
        productions.add_exchanges(cost=10, used=True, ex=[exchange0, exchange1])
        productions.add_exchanges(cost=10, used=False, ex=[exchange2, exchange3, exchange4])

        exchanges = LedgerExchange()
        exchanges.add_all([exchange0, exchange1, exchange2, exchange3, exchange4], 'export')

        state = State(name='fr',
                      consumptions=LedgerConsumption(),
                      productions=productions,
                      borders=LedgerBorder(), rac=0, cost=0)
        state.exchanges = exchanges

        # Expected
        exp_productions = LedgerProduction(uuid_generate=lambda: None)
        exp_productions.add_exchanges(cost=10, used=True,
                                     ex=[Exchange(id=0, production_type='solar', quantity=1, path_node=['fr'])])
        exp_productions.add_exchanges(cost=10, used=True,
                                     ex=[Exchange(id=1, production_type='solar', quantity=1, path_node=['fr'])])

        exp_exchanges = LedgerExchange()
        exp_exchanges.add_all([Exchange(id=0, production_type='solar', quantity=1, path_node=['fr']),
                               Exchange(id=1, production_type='solar', quantity=1, path_node=['fr'])], 'export')

        exp_state = State(name='fr',
                          consumptions=LedgerConsumption(),
                          productions=exp_productions,
                          borders=LedgerBorder(), rac=0, cost=0)
        exp_state.exchanges = exp_exchanges

        handler = CancelUselessImportationHandler(next=ReturnHandler(), params=HandlerParameter(tell=tell_mock))
        new_state, _ = handler.execute(state)

        # Test
        self.assertEqual(exp_state, new_state)

        tell_mock.assert_has_calls(
            [call(to='fr', mes=ConsumerCanceledExchange(exchanges=[exchange2], path_node=['fr'])),
             call(to='be', mes=ConsumerCanceledExchange(exchanges=[exchange3, exchange4], path_node=['be']))])


class TestProposeFreeProduction(unittest.TestCase):

    def test_execute_enough_border(self):
        # Create mock
        tell_mock = MagicMock()
        params = HandlerParameter(tell=tell_mock)

        mock_uuid = MockUUID()

        # Input
        productions = LedgerProduction(uuid_generate=mock_uuid.generate)
        productions.add_production(type='solar', cost=10, quantity=10, used=True)
        productions.add_production(type='nuclear', cost=20, quantity=10, used=True)
        productions.add_production(type='nuclear', cost=20, quantity=10, used=False)

        borders = LedgerBorder()
        borders.add(dest='be', cost=10, quantity=10)  # Has enough quantity

        exchanges = LedgerExchange()
        exchanges.add(Exchange(id=1, quantity=5, production_type='nuclear', path_node=['be']), type='export')

        state = State(name='fr', consumptions=LedgerConsumption(), borders=borders, productions=productions, cost=0,
                      rac=0)
        state.exchanges = exchanges

        handler = ProposeFreeProductionHandler(next=ReturnHandler(), params=params)
        new_sate, _ = handler.execute(state)

        self.assertEqual(state, new_sate)
        tell_mock.assert_called_with(to='be', mes=Proposal(production_type='nuclear', cost=30, quantity=5, path_node=['fr']))

    def test_execute_saturation_border(self):
        # Create mock
        tell_mock = MagicMock()
        params = HandlerParameter(tell=tell_mock)

        mock_uuid = MockUUID()

        # Input
        productions = LedgerProduction(uuid_generate=mock_uuid.generate)
        productions.add_production(type='solar', cost=10, quantity=10, used=True)
        productions.add_production(type='nuclear', cost=20, quantity=10, used=True)
        productions.add_production(type='nuclear', cost=20, quantity=10, used=False)

        borders = LedgerBorder()
        borders.add(dest='be', cost=10, quantity=7)  # Not enough quantity

        exchanges = LedgerExchange()
        exchanges.add(Exchange(id=1, quantity=5, production_type='nuclear', path_node=['be']), 'export')

        state = State(name='fr', consumptions=LedgerConsumption(), borders=borders, productions=productions, cost=0,
                      rac=0)
        state.exchanges = exchanges

        handler = ProposeFreeProductionHandler(next=ReturnHandler(), params=params)
        new_sate, _ = handler.execute(state)

        self.assertEqual(state, new_sate)
        tell_mock.assert_called_with(to='be', mes=Proposal(production_type='nuclear', cost=30, quantity=2, path_node=['fr']))


class TestForwardMessageHandler(unittest.TestCase):
    def test_execute(self):
        # Create mock
        tell_mock = MagicMock()
        params = HandlerParameter(tell=tell_mock)

        # Input
        borders = LedgerBorder()
        borders.add(dest='be', cost=2, quantity=10)
        borders.add(dest='de', cost=4, quantity=5)
        borders.add(dest='uk', cost=10, quantity=10)

        state = State(name='fr', consumptions=None, borders=borders, productions=None, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add(ex=Exchange(quantity=5, id=0, production_type='solar', path_node=['uk', 'fr']), type='transfer')

        prop = Proposal(production_type='solar', cost=10, quantity=7, path_node=['it'])

        # Expected
        be_prop = Proposal(production_type='solar', cost=12, quantity=7, path_node=['fr', 'it'])
        de_prop = Proposal(production_type='solar', cost=14, quantity=5, path_node=['fr', 'it'])
        uk_prop = Proposal(production_type='solar', cost=20, quantity=5, path_node=['fr', 'it'])

        # Test
        handler = ForwardMessageHandler(next=ReturnHandler(), params=params)
        res_state, res_message = handler.execute(state=state, message=prop)

        tell_mock.assert_has_calls(
            [call(to='be', mes=be_prop),
             call(to='de', mes=de_prop),
             call(to='uk', mes=uk_prop)])


class TestSaveExchangeHandler(unittest.TestCase):
    def test_trim_path(self):
        self.assertEqual(['be'], SaveExchangeHandler.trim_path('fr', ['it', 'fr', 'be']))
        self.assertEqual(['be'], SaveExchangeHandler.trim_path('fr', ['fr', 'be']))
        self.assertEqual(['be'], SaveExchangeHandler.trim_path('fr', ['be']))
        self.assertEqual([], SaveExchangeHandler.trim_path('be', ['be']))


class TestCancelExportationHandler(unittest.TestCase):
    def test_execute(self):
        # Create mock
        params = HandlerParameter()

        # Input
        ex_cancel = Exchange(id=10, production_type='solar', quantity=10, path_node=['fr', 'be', 'de'])
        ex_keep = Exchange(id=5, production_type='solar', quantity=10, path_node=['fr', 'be', 'de'])

        state = State(name='fr', consumptions=LedgerConsumption(),
                      borders=LedgerBorder(), productions=LedgerProduction(), rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add_all([ex_cancel, ex_keep], 'export')

        message = ConsumerCanceledExchange(path_node=['fr', 'be'], exchanges=[ex_cancel])

        # Expected
        state_exp = State(name='fr', consumptions=LedgerConsumption(),
                          borders=LedgerBorder(), productions=LedgerProduction(), rac=0, cost=0)
        state_exp.exchanges = LedgerExchange()
        state_exp.exchanges.add(ex_keep, 'export')

        # Test
        handler = CancelExportationHandler(next=ReturnHandler(), params=params)
        res, _ = handler.execute(state=state, message=message)

        self.assertEqual(state_exp, res)


class TestBackwardMessageHandler(unittest.TestCase):
    def test_execute_tell(self):
        # Create mock
        tell_mock = MagicMock()
        params = HandlerParameter(tell=tell_mock)

        # Input
        message = ProposalOffer(production_type='solar', cost=10, quantity=10,
                                path_node=['fr', 'de'], return_path_node=['de', 'fr'])
        state = State(name='fr', consumptions=None, borders=None, productions=None, rac=0, cost=0)

        # Expected
        expected = ProposalOffer(production_type='solar', cost=10, quantity=10,
                                path_node=['de'], return_path_node=['de', 'fr'])

        # Test
        handler = BackwardMessageHandler(after_backward=ReturnHandler(), on_resume=ReturnHandler(), params=params, type='tell')
        res, _ = handler.execute(state=state, message=message)

        self.assertEqual(state, res)
        tell_mock.assert_called_with(to='de', mes=expected)

    def test_execute_ask(self):
        # Expected
        return_expected = [Exchange(quantity=10, id=1, production_type='solar', path_node=['be', 'fr', 'de'])]
        ask_expected = ProposalOffer(production_type='solar', cost=10, quantity=10,
                                path_node=['de'], return_path_node=['de', 'fr'])

        # Create mock
        ask_mock = MagicMock(return_value=return_expected)
        params = HandlerParameter(ask=ask_mock)

        # Input
        message = ProposalOffer(production_type='solar', cost=10, quantity=10,
                                path_node=['fr', 'de'], return_path_node=['de', 'fr'])
        state = State(name='fr', consumptions=None, borders=None, productions=None, rac=0, cost=0)

        # Test
        handler = BackwardMessageHandler(after_backward=ReturnHandler(), on_resume=ReturnHandler(), params=params, type='ask')
        res_state, res_message = handler.execute(state=state, message=message)

        self.assertEqual(state, res_state)
        self.assertEqual(return_expected, res_message)
        ask_mock.assert_called_with(to='de', mes=ask_expected)

    def test_execute_resume(self):
        # Create mock
        params = HandlerParameter()

        # Input
        message = ProposalOffer(production_type='solar', cost=10, quantity=10,
                                path_node=['de'], return_path_node=['fr', 'be'])
        state = State(name='de', consumptions=None, borders=None, productions=None, rac=0, cost=0)

        # Test
        handler = BackwardMessageHandler(after_backward=ReturnHandler(), on_resume=ReturnHandler(), params=params, type='ask')
        res_state, res_message = handler.execute(state=state, message=message)

        self.assertEqual(state, res_state)
        self.assertEqual(message, res_message)


class TestAcceptAvailableExchangeHandler(unittest.TestCase):
    def test_execute(self):
        # Input
        uuid_mock = MockUUID()
        params = HandlerParameter(uuid_generate=uuid_mock.generate)

        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(cost=10, quantity=10, type='nuclear', used=False)

        state = State(name='fr', consumptions=None, productions=productions, borders=None, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add_all([Exchange(quantity=1, id=1, production_type='nuclear', path_node=['be']),
                                 Exchange(quantity=1, id=2, production_type='nuclear', path_node=['be']),
                                 Exchange(quantity=1, id=3, production_type='nuclear', path_node=['be']),
                                 Exchange(quantity=1, id=4, production_type='nuclear', path_node=['be']),
                                 Exchange(quantity=1, id=5, production_type='nuclear', path_node=['be'])], 'export')

        offer = ProposalOffer(production_type='nuclear', cost=12, quantity=10, path_node=['fr'], return_path_node=['de'])

        # Expected
        response_exp = [Exchange(quantity=1, id=6, production_type='nuclear', path_node=['de']),
                        Exchange(quantity=1, id=7, production_type='nuclear', path_node=['de']),
                        Exchange(quantity=1, id=8, production_type='nuclear', path_node=['de']),
                        Exchange(quantity=1, id=9, production_type='nuclear', path_node=['de']),
                        Exchange(quantity=1, id=10, production_type='nuclear', path_node=['de'])]

        # Test
        handler = AcceptExchangeHandler(next=ReturnHandler(), params=params)
        state_res, response_res = handler.execute(state, offer)

        self.assertEqual(state, state_res)
        self.assertEqual(response_exp, response_res)


class TestCompareNewProduction(unittest.TestCase):
    def test_execute_on_expensive(self):
        # Input
        params = HandlerParameter()

        consumptions = LedgerConsumption()
        consumptions.add(type='load', cost=10**3, quantity=2)

        uuid_mock = MockUUID()
        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(type='solar', cost=10, quantity=2, used=True)

        state = State(name='fr', consumptions=consumptions, borders=None, productions=productions, rac=0, cost=20)
        proposal = Proposal(production_type='nuclear', cost=12, quantity=10, path_node=['it'])

        # Create mock

        for_prod_useless = ReturnHandler()
        for_prod_useless.execute = MagicMock(return_value=(state, proposal))

        for_prod_useful = ReturnHandler()
        for_prod_useful.execute = MagicMock()

        # Test
        handler = CompareNewProduction(for_prod_useless=for_prod_useless, for_prod_useful=for_prod_useful, params=params)
        state_res, mes_res = handler.execute(state, proposal)

        self.assertEqual(state, state_res)
        self.assertEqual(proposal, mes_res)

    def test_execute_on_cheaper(self):
        # Input
        params = HandlerParameter()

        consumptions = LedgerConsumption()
        consumptions.add(type='load', cost=10**3, quantity=2)

        uuid_mock = MockUUID()
        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(type='solar', cost=15, quantity=2, used=True)

        state = State(name='fr', consumptions=consumptions, borders=None, productions=productions, rac=0, cost=30)
        proposal = Proposal(production_type='nuclear', cost=12, quantity=4, path_node=['it'])

        # Create mock
        for_prod_useless = ReturnHandler()
        for_prod_useless.execute = MagicMock()

        for_prod_useful = ReturnHandler()
        for_prod_useful.execute = MagicMock(return_value=(state, proposal))

        # Expected
        expected_prop = Proposal(production_type='nuclear', cost=12, quantity=2, path_node=['it'])

        # Test
        handler = CompareNewProduction(for_prod_useless=for_prod_useless, for_prod_useful=for_prod_useful,
                                       params=params)
        state_res, mes_res = handler.execute(state, proposal)

        self.assertEqual(state, state_res)
        self.assertEqual(proposal, mes_res)
        for_prod_useless.execute.assert_called_with(state=state, message=expected_prop)


class TestMakeOfferHandler(unittest.TestCase):
    def test_execute(self):
        # Mock
        exs_return = [Exchange(quantity=1, id=0, production_type='solar', path_node=['it'])]
        ask_mock = MagicMock(return_value=exs_return)
        params = HandlerParameter(ask=ask_mock)

        # Input
        state = State(name='fr', consumptions=None, borders=None, productions=LedgerProduction(), rac=0, cost=0)
        proposal = Proposal(production_type='solar', cost=10, quantity=1, path_node=['it'])

        # Expected
        offer_expected = ProposalOffer(production_type='solar', cost=10, quantity=1, path_node=['it'], return_path_node=['fr'])

        exs_expected = [Exchange(quantity=1, id=0, production_type='solar', path_node=['it'])]

        state_expected = deepcopy(state)
        state_expected.productions.add_exchanges(cost=10, ex=exs_expected)

        # Test
        handler = MakerOfferHandler(next=ReturnHandler(), params=params)
        res_state, res_mes = handler.execute(deepcopy(state), deepcopy(proposal))

        self.assertEqual(state_expected, res_state)
        self.assertEqual(exs_expected, res_mes)
        ask_mock.assert_called_with(to='it', mes=offer_expected)


class TestCheckOfferBorderCapacityHandler(unittest.TestCase):
    def test_execute_available(self):
        # Input
        params = HandlerParameter()

        borders = LedgerBorder()
        borders.add(dest='be', cost=10, quantity=10)

        state = State(name='fr', consumptions=None, borders=borders, productions=None, rac=0, cost=0)
        state.exchanges = LedgerExchange()
        state.exchanges.add(Exchange(id=0, production_type='solar', quantity=5, path_node=['be']), type='export')

        offer = ProposalOffer(production_type='solar', cost=10, quantity=10, path_node=['fr', 'it'], return_path_node=['fr', 'be'])

        # Expected
        offer_exp = ProposalOffer(production_type='solar', cost=10, quantity=5, path_node=['fr', 'it'], return_path_node=['fr', 'be'])

        handler = CheckOfferBorderCapacityHandler(next=ReturnHandler(), params=params)
        res_state, res_message = handler.execute(state=state, message=offer)

        self.assertEqual(state, res_state)
        self.assertEqual(offer_exp, res_message)


class TestAdequacyHandler(unittest.TestCase):
    def test_execute(self):
        # Input
        params = HandlerParameter()

        consumptions = LedgerConsumption()
        consumptions.add(type='load', cost=100, quantity=10)

        uuid_mock = MockUUID()
        productions = LedgerProduction(uuid_generate=uuid_mock.generate)
        productions.add_production(type='solar', cost=10, quantity=4)
        productions.add_production(type='nuclear', cost=20, quantity=4)

        state = State(name='fr', consumptions=consumptions, borders=None, productions=productions, rac=0, cost=0)

        # Expected
        consumptions = LedgerConsumption()
        consumptions.ledger = pd.DataFrame({'cost': [100], 'quantity': [10]}, index=['load'])

        productions = LedgerProduction()
        productions.ledger = pd.DataFrame({'cost': [10] * 4 + [20] * 4,
                                           'quantity': [4] * 8,
                                           'type': ['solar'] * 4 + ['nuclear'] * 4,
                                           'path_node': [None] * 8},
                                          index=[1, 2, 3, 4, 5, 6, 7, 8])
        expected = State(name='fr', consumptions=consumptions, borders=None, productions=productions, rac=-2, cost=320)

        handler = AdequacyHandler(next=ReturnHandler(), params=params)
        new_state, _ = handler.execute(state=state, message=None)


class MockUUID:
    def __init__(self):
        self.inc = 0

    def generate(self):
        self.inc += 1
        return self.inc



import unittest
import pandas as pd

from hadar.aggregator.result import Index, TimeIndex, ResultAggregator, NodeIndex, TypeIndex, SrcIndex, DestIndex
from hadar.solver.input import *
from hadar.solver.output import *


class TestIndex(unittest.TestCase):

    def test_no_parameters(self):
        self.assertEqual(True, Index(column='i').all)

    def test_on_element(self):
        i = Index(column='i', index='fr')
        self.assertEqual(False, i.all)
        self.assertEqual(['fr'], i.index)

    def test_list(self):
        i = Index(column='i', index=['fr', 'be'])
        self.assertEqual(False, i.all)
        self.assertEqual(['fr', 'be'], i.index)

    def test_filter(self):
        i = Index(column='i', index=['fr', 'be'])
        df = pd.DataFrame(data={'i': ['it', 'fr', 'fr', 'be', 'de', 'it', 'be'],
                                'a': [0, 1, 2, 3, 4, 5, 6]})

        exp = pd.Series(data=[False, True, True, True, False, False, True], index=[0, 1, 2, 3, 4, 5, 6], name='i')

        pd.testing.assert_series_equal(exp, i.filter(df))


class TestTimeIndex(unittest.TestCase):

    def test_wrong_range(self):
        self.assertRaises(ValueError, lambda: TimeIndex(start=56))
        self.assertRaises(ValueError, lambda: TimeIndex(end=23))

    def test_range(self):
        i = TimeIndex(start=2, end=6)
        self.assertEqual(False, i.all)
        self.assertEqual([2, 3, 4, 5], i.index)

    def test_list(self):
        i = TimeIndex(index=[2, 6])
        self.assertEqual(False, i.all)
        self.assertEqual([2, 6], i.index)


class TestAggregator(unittest.TestCase):
    def setUp(self) -> None:
        self.study = Study(['a', 'b', 'c']) \
            .add_on_node('a', data=Consumption(cost=10 ** 3, quantity=[120, 12], type='load')) \
            .add_on_node('a', data=Consumption(cost=10 ** 3, quantity=[130, 13], type='car')) \
            .add_on_node('a', data=Production(cost=10, quantity=[130, 13], type='prod')) \
            .add_on_node('b', data=Consumption(cost=10 ** 3, quantity=[120, 12], type='load')) \
            .add_on_node('b', data=Production(cost=20, quantity=[110, 11], type='prod')) \
            .add_on_node('b', data=Production(cost=20, quantity=[120, 12], type='nuclear')) \
            .add_border(src='a', dest='b', quantity=[110, 11], cost=2) \
            .add_border(src='a', dest='c', quantity=[120, 12], cost=2)

        out = {
            'a': OutputNode(consumptions=[OutputConsumption(cost=10 ** 3, quantity=[20, 2], type='load'),
                                          OutputConsumption(cost=10 ** 3, quantity=[30, 3], type='car')],
                            productions=[OutputProduction(cost=10, quantity=[30, 3], type='prod')],
                            borders=[OutputBorder(dest='b', quantity=[10, 1], cost=2),
                                     OutputBorder(dest='c', quantity=[20, 2], cost=2)]),

            'b': OutputNode(consumptions=[OutputConsumption(cost=10 ** 3, quantity=[20, 2], type='load')],
                            productions=[OutputProduction(cost=20, quantity=[10, 1], type='prod'),
                                         OutputProduction(cost=20, quantity=[20, 2], type='nuclear')],
                            borders=[])
        }

        self.result = Result(nodes=out)

    def test_build_consumption(self):
        # Expected
        exp = pd.DataFrame(data={'cost': [10 ** 3] * 6,
                                 'asked': [120, 12, 130, 13, 120, 12],
                                 'given': [20, 2, 30, 3, 20, 2],
                                 'type': ['load', 'load', 'car', 'car', 'load', 'load'],
                                 'node': ['a', 'a', 'a', 'a', 'b', 'b'],
                                 't': [0, 1, 0, 1, 0, 1]}, dtype=float)

        agg = ResultAggregator(study=self.study, result=self.result)
        cons = agg._build_consumption()

        pd.testing.assert_frame_equal(exp, cons)

    def test_build_production(self):
        # Expected
        exp = pd.DataFrame(data={'cost': [10, 10, 20, 20, 20, 20],
                                 'avail': [130, 13, 110, 11, 120, 12],
                                 'used': [30, 3, 10, 1, 20, 2],
                                 'type': ['prod', 'prod', 'prod', 'prod', 'nuclear', 'nuclear'],
                                 'node': ['a', 'a', 'b', 'b', 'b', 'b'],
                                 't': [0, 1, 0, 1, 0, 1]}, dtype=float)

        agg = ResultAggregator(study=self.study, result=self.result)
        prod = agg._build_production()

        pd.testing.assert_frame_equal(exp, prod)

    def test_build_border(self):
        # Expected
        exp = pd.DataFrame(data={'cost': [2, 2, 2, 2],
                                 'avail': [110, 11, 120, 12],
                                 'used': [10, 1, 20, 2],
                                 'src': ['a', 'a', 'a', 'a'],
                                 'dest': ['b', 'b', 'c', 'c'],
                                 't': [0, 1, 0, 1]}, dtype=float)

        agg = ResultAggregator(study=self.study, result=self.result)
        border = agg._build_border()

        pd.testing.assert_frame_equal(exp, border)

    def test_aggregate_cons(self):
        # Expected
        index = pd.Index(data=[0, 1], dtype=float, name='t')
        exp_cons = pd.DataFrame(data={'asked': [120, 12],
                                      'cost': [10 ** 3] * 2,
                                      'given': [20, 2]}, dtype=float, index=index)

        agg = ResultAggregator(study=self.study, result=self.result)
        cons = agg.agg_cons(i0=NodeIndex(index='a'), i1=TypeIndex(index='load'), i2=TimeIndex())

        pd.testing.assert_frame_equal(exp_cons, cons)

    def test_aggregate_prod(self):
        # Expected
        index = pd.MultiIndex.from_tuples((('a', 'prod', 0.0), ('a', 'prod', 1.0), ('b', 'prod', 0.0), ('b', 'prod', 1.0)),
                                          names=['node', 'type', 't'], )
        exp_cons = pd.DataFrame(data={'avail': [130, 13, 110, 11],
                                      'cost': [10, 10, 20, 20],
                                      'used': [30, 3, 10, 1]}, dtype=float, index=index)

        agg = ResultAggregator(study=self.study, result=self.result)
        cons = agg.agg_prod(i0=NodeIndex(index=['a', 'b']), i1=TypeIndex(index='prod'), i2=TimeIndex())

        pd.testing.assert_frame_equal(exp_cons, cons)

    def test_aggregate_border(self):
        # Expected
        index = pd.MultiIndex.from_tuples((('b', 0.0), ('b', 1.0), ('c', 0.0), ('c', 1.0)),
                                          names=['dest', 't'], )
        exp_cons = pd.DataFrame(data={'avail': [110, 11, 120, 12],
                                      'cost': [2, 2, 2, 2],
                                      'used': [10, 1, 20, 2]}, dtype=float, index=index)

        agg = ResultAggregator(study=self.study, result=self.result)
        cons = agg.agg_border(i0=SrcIndex(index=['a']), i1=DestIndex(index=['b', 'c']), i2=TimeIndex())

        pd.testing.assert_frame_equal(exp_cons, cons)

    def test_get_elements_inide(self):
        agg = ResultAggregator(study=self.study, result=self.result)
        self.assertEqual((2, 1, 2), agg.get_elements_inside('a'))
        self.assertEqual((1, 2, 0), agg.get_elements_inside('b'))

    def test_balance(self):
        agg = ResultAggregator(study=self.study, result=self.result)
        np.testing.assert_array_equal([30, 3], agg.get_balance(node='a'))
        np.testing.assert_array_equal([-10, -1], agg.get_balance(node='b'))

    def test_cost(self):
        agg = ResultAggregator(study=self.study, result=self.result)
        np.testing.assert_array_equal([200360, 20036], agg.get_cost(node='a'))
        np.testing.assert_array_equal([100600, 10060], agg.get_cost(node='b'))





import unittest
from hadar.solver.input import *

class TestNetwork(unittest.TestCase):
    def test_create_study(self):
        c = Consumption(type='load', cost=20, quantity=[10])
        p = Production(type='nuclear', cost=20, quantity=[10])
        b = Border(dest='a', cost=20, quantity=[10])

        study = Study(['a', 'b']) \
            .add_on_node(node='a', data=c) \
            .add_on_node(node='a', data=p) \
            .add_border(src='b', dest='a', cost=20, quantity=[10])

        self.assertEqual(c, study.nodes['a'].consumptions[0])
        self.assertEqual(p, study.nodes['a'].productions[0])
        self.assertEqual(b, study.nodes['b'].borders[0])
        self.assertEqual(1, study.horizon)

    def test_wrong_node_list(self):
        def test():
            study = Study(node_names=['fr', 'be', 'de', 'be'])
        self.assertRaises(ValueError, test)

    def test_wrong_production_cost(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Production(type='solar', cost=-1, quantity=[10]))

        self.assertRaises(ValueError, test)

    def test_wrong_production_quantity(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Production(type='solar', cost=10, quantity=[-1]))

        self.assertRaises(ValueError, test)

    def test_wrong_production_type(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Production(type='solar', cost=10, quantity=[10])) \
                .add_on_node(node='fr', data=Production(type='solar', cost=10, quantity=[10]))

        self.assertRaises(ValueError, test)

    def test_wrong_consumption_cost(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Consumption(type='load', cost=-10, quantity=[10]))

        self.assertRaises(ValueError, test)

    def test_wrong_consumption_quantity(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Consumption(type='load', cost=10, quantity=[-10]))

        self.assertRaises(ValueError, test)

    def test_wrong_consumption_type(self):
        def test():
            study = Study(node_names=['fr']) \
                .add_on_node(node='fr', data=Consumption(type='load', cost=10, quantity=[10])) \
                .add_on_node(node='fr', data=Consumption(type='load', cost=10, quantity=[10]))

        self.assertRaises(ValueError, test)

    def test_wrong_border_cost(self):
        def test():
            study = Study(node_names=['fr', 'be']) \
                .add_border(src='fr', dest='be', cost=-10, quantity=[10])

        self.assertRaises(ValueError, test)

    def test_wrong_border_quantity(self):
        def test():
            study = Study(node_names=['fr', 'be']) \
                .add_border(src='fr', dest='be', cost=10, quantity=[-10])

        self.assertRaises(ValueError, test)

    def test_wrong_border_dest_not_node(self):
        def test():
            study = Study(node_names=['fr', 'be']) \
                .add_border(src='fr', dest='it', cost=10, quantity=[10])

        self.assertRaises(ValueError, test)

    def test_wrong_border_dest_not_unique(self):
        def test():
            study = Study(node_names=['fr', 'be']) \
                .add_border(src='fr', dest='be', cost=10, quantity=[10]) \
                .add_border(src='fr', dest='be', cost=10, quantity=[10])
        self.assertRaises(ValueError, test)



import unittest

from hadar.solver.input import *
from hadar.solver.output import *
from hadar.solver.study import solve
from tests.utils import assert_study


class TestSolver(unittest.TestCase):

    def test_merit_order(self):
        """
        Capacity
        |          A           |
        | load: 30             |
        | solar: 10 @ 10       |
        | nuclear: 15 @ 20     |
        | gas: 10 @ 30         |


        Adequacy
        |          A           |
        | load: 30             |
        | solar: 10            |
        | nuclear: 15          |
        | gas: 5               |
        :return:
        """
        study = Study(['a']) \
            .add_on_node(node='a', data=Consumption(type='load', cost=10 ** 6, quantity=[30, 6])) \
            .add_on_node(node='a', data=Production(type='nuclear', cost=20, quantity=[15, 3])) \
            .add_on_node(node='a', data=Production(type='solar', cost=10, quantity=[10, 2])) \
            .add_on_node(node='a', data=Production(type='oil', cost=30, quantity=[10, 2]))

        nodes_expected = dict()
        nodes_expected['a'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[30, 6], type='load')],
                                         productions=[
                                             OutputProduction(type='nuclear', cost=20, quantity=[15, 3]),
                                             OutputProduction(type='solar', cost=10, quantity=[10, 2]),
                                             OutputProduction(type='oil', cost=30, quantity=[5, 1])],
                                         borders=[])

        res = solve(study, kind='lp')
        assert_study(self, Result(nodes_expected), res)

    def test_exchange_two_nodes(self):
        # Input
        study = Study(['a', 'b']) \
            .add_on_node('a', data=Consumption(cost=10 ** 6, quantity=[20, 200], type='load')) \
            .add_on_node('a', data=Production(cost=10, quantity=[30, 300], type='prod')) \
            .add_on_node('b', data=Consumption(cost=10 ** 6, quantity=[20, 200], type='load')) \
            .add_on_node('b', data=Production(cost=20, quantity=[10, 100], type='prod')) \
            .add_border(src='a', dest='b', quantity=[10, 100], cost=2)

        nodes_expected = {}
        nodes_expected['a'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[20, 200], type='load')],
                                         productions=[OutputProduction(cost=10, quantity=[30, 300], type='prod')],
                                         borders=[OutputBorder(dest='b', quantity=[10, 100], cost=2)])

        nodes_expected['b'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[20, 200], type='load')],
                                         productions=[OutputProduction(cost=20, quantity=[10, 100], type='prod')],
                                         borders=[])


        res = solve(study)
        assert_study(self, Result(nodes_expected), res)


    def test_exchange_two_concurrent_nodes(self):
        """
        Capacity
        |          A           | --------> |          B           |
        | load: 10             |  10       | load: 10             |
        | nuclear: 30 @ 10     | ----.     | nuclear: 10 @ 20     |
                                  10 |
                                     .---> |          C           |
                                           | load: 10             |
                                           | nuclear: 10 @ 20     |

        Adequacy
        |          A           | --------> |          B           |
        | load: 10             | 10        | load: 10             |
        | nuclear: 30          | ----.     | nuclear: 0           |
                                 10  |
                                     .---> |          C           |
                                           | load: 10             |
                                           | nuclear: 0           |
        :return:
        """
        study = Study(node_names=['a', 'b', 'c']) \
            .add_on_node('a', data=Consumption(cost=10 ** 6, quantity=[10], type='load')) \
            .add_on_node('a', data=Production(cost=10, quantity=[30], type='nuclear')) \
            .add_on_node('b', data=Consumption(cost=10 ** 6, quantity=[10], type='load')) \
            .add_on_node('b', data=Production(cost=20, quantity=[10], type='nuclear')) \
            .add_on_node('c', data=Consumption(cost=10 ** 6, quantity=[10], type='load')) \
            .add_on_node('c', data=Production(cost=20, quantity=[10], type='nuclear')) \
            .add_border(src='a', dest='b', quantity=[20], cost=2) \
            .add_border(src='a', dest='c', quantity=[20], cost=2)

        nodes_expected = {}
        nodes_expected['a'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         productions=[OutputProduction(cost=10, quantity=[30], type='nuclear')],
                                         borders=[OutputBorder(dest='b', quantity=[10], cost=2),
                                                  OutputBorder(dest='c', quantity=[10], cost=2)])

        nodes_expected['b'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         productions=[OutputProduction(cost=20, quantity=[0], type='nuclear')],
                                         borders=[])

        nodes_expected['c'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         productions=[OutputProduction(cost=20, quantity=[0], type='nuclear')],
                                         borders=[])

        res = solve(study, kind='lp')

        assert_study(self, Result(nodes_expected), res)

    def test_exchange_border_saturation(self):
        """
        Capacity
        |           A           | ---->  |           B           | ---->  |           C           |
        | nuclear: 30 @ 10      | 20     |  load: 10             | 15     | load: 20              |


        Adequacy
        |           A           | ---->  |           B           | ---->  |           C           |
        | nuclear: 20           | 20     |  load: 10             | 10     | load: 10              |

        :return:
        """
        study = Study(node_names=['a', 'b', 'c']) \
            .add_on_node('a', data=Production(cost=10, quantity=[30], type='nuclear')) \
            .add_on_node('b', data=Consumption(cost=10 ** 6, quantity=[10], type='load')) \
            .add_on_node('c', data=Consumption(cost=10 ** 6, quantity=[20], type='load')) \
            .add_border(src='a', dest='b', quantity=[20], cost=2) \
            .add_border(src='b', dest='c', quantity=[15], cost=2)

        nodes_expected = {}
        nodes_expected['a'] = OutputNode(productions=[OutputProduction(cost=10, quantity=[20], type='nuclear')],
                                         borders=[OutputBorder(dest='b', quantity=[20], cost=2)],
                                         consumptions=[])

        nodes_expected['b'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         borders=[OutputBorder(dest='c', quantity=[10], cost=2)],
                                         productions=[])

        nodes_expected['c'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         productions=[],
                                         borders=[])

        res = solve(study, kind='lp')

        assert_study(self, Result(nodes_expected), res)

    def test_consumer_cancel_exchange(self):
        """
        Capacity
        |           A           | ---->  |           B           | ---->  |           C           |
        | load: 10              |  20    | load: 5               |  20    | load: 20              |
        | nuclear: 20 @ 10      |        | nuclear: 15 @ 20      |        | nuclear: 10 @ 10      |


        Adequacy
        |           A           | ---->  |           B           | ---->  |           C           |
        | load: 10              | 10     | load: 5               | 10     | load: 20              |
        | nuclear: 20           |        | nuclear: 5            |        | nuclear: 10           |


        :return:
        """
        study = Study(node_names=['a', 'b', 'c']) \
            .add_on_node('a', data=Consumption(cost=10 ** 6, quantity=[10], type='load')) \
            .add_on_node('a', data=Production(cost=10, quantity=[20], type='nuclear')) \
            .add_on_node('b', data=Consumption(cost=10 ** 6, quantity=[5], type='load')) \
            .add_on_node('b', data=Production(cost=20, quantity=[15], type='nuclear')) \
            .add_on_node('c', data=Consumption(cost=10 ** 6, quantity=[20], type='load')) \
            .add_on_node('c', data=Production(cost=10, quantity=[10], type='nuclear')) \
            .add_border(src='a', dest='b', quantity=[20], cost=2) \
            .add_border(src='b', dest='c', quantity=[20], cost=2)

        nodes_expected = {}
        nodes_expected['a'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[10], type='load')],
                                         productions=[OutputProduction(cost=10, quantity=[20], type='nuclear')],
                                         borders=[OutputBorder(dest='b', quantity=[10], cost=2)])

        nodes_expected['b'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[5], type='load')],
                                         productions=[OutputProduction(cost=20, quantity=[5], type='nuclear')],
                                         borders=[OutputBorder(dest='c', quantity=[10], cost=2)])

        nodes_expected['c'] = OutputNode(consumptions=[OutputConsumption(cost=10 ** 6, quantity=[20], type='load')],
                                         productions=[OutputProduction(cost=10, quantity=[10], type='nuclear')],
                                         borders=[])

        res = solve(study, kind='lp')

        assert_study(self, Result(nodes_expected), res)



import logging
import time

from pykka import ActorRegistry

from hadar.solver.actor.actor import Dispatcher, Result, Waiter
from hadar.solver.input import *
from hadar.solver.actor.domain.message import Start, Next


logger = logging.getLogger(__name__)


def _create_dispatcher(name: str, node: InputNode, waiter: Waiter) -> Dispatcher:
    return Dispatcher.start(name=name, waiter=waiter, input=node)


def solve(study: Study) -> Result:
    """
    Solve adequacy by behaviour simulation with actor pattern.

    :param study: study to resolve
    :return: result
    """
    waiter = Waiter(wait_ms=100)

    dispatchers = [_create_dispatcher(name, node, waiter) for name, node in study.nodes.items()]
    for d in dispatchers:
        d.tell(Start())

    logger.info('Agents stated')
    waiter.wait()

    nodes = {}
    for d in dispatchers:
        name, node = d.ask(Next())
        nodes[name] = node
    logger.info('Agents stopped')

    #for d in dispatchers:
    #    plot(d.ask(Snapshot()))

    ActorRegistry.stop_all()
    return Result(nodes=nodes)


import unittest
import hashlib
import pandas as pd
import numpy as np

import plotly.graph_objects as go
from plotly.offline.offline import plot

from hadar.solver.input import *
from hadar.solver.output import *
from hadar.solver.study import solve
from hadar.aggregator.result import *
from hadar.viewer.html import HTMLPlotting


class TestHTMLPlotting(unittest.TestCase):
    def setUp(self) -> None:
        self.study = Study(['a', 'b']) \
            .add_on_node('a', data=Consumption(cost=10 ** 6, quantity=[20, 2], type='load')) \
            .add_on_node('a', data=Consumption(cost=10 ** 6, quantity=[30, 3], type='car')) \
            .add_on_node('a', data=Production(cost=10, quantity=[60, 3], type='prod')) \
            .add_on_node('b', data=Consumption(cost=10 ** 6, quantity=[40, 2], type='load')) \
            .add_on_node('b', data=Production(cost=20, quantity=[10, 1], type='prod')) \
            .add_on_node('b', data=Production(cost=20, quantity=[20, 2], type='nuclear')) \
            .add_border(src='a', dest='b', quantity=[10, 1], cost=2)

        self.result = solve(study=self.study)

        self.agg = ResultAggregator(self.study, self.result)
        self.plot = HTMLPlotting(agg=self.agg, unit_symbol='MW', time_start='2020-02-01', time_end='2020-02-02',
                                 node_coord={'a': [2.33, 48.86], 'b': [4.38, 50.83]})

        self.hash = hashlib.sha3_256()

    def test_stack(self):
        fig = self.plot.stack('a')
        self.assert_fig('6ef4b03c8b7e95b00e483fbc58e084486b293d06', fig)

    def test_map_exchanges(self):
        fig = self.plot.exchanges_map(0)
        self.assert_fig('9aa34f28665ea9e6766b271ffbc677d3cda6810b', fig)

    def assert_fig(self, expected: str, fig: go.Figure):
        h = hashlib.sha1()
        h.update(TestHTMLPlotting.get_html(fig).encode('ascii'))
        self.assertEqual(expected, h.hexdigest())

    @staticmethod
    def get_html(fig: go.Figure):
        html = plot(fig, include_plotlyjs=False, include_mathjax=False, output_type='div')
        # plotly use a random id. We need to extract it and replace it by constant
        # uuid can be find at ... <div id="xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx" class="plotly-graph-div" ...
        uuid = html.split('" class="plotly-graph-div"')[0][-36:]
        return html.replace(uuid, '#####')


from ortools.linear_solver.pywraplp import Solver

from hadar.solver.input import DTO


class MockNumVar(DTO):
    def __init__(self, min: float, max: float, name: str):
        self.min = min
        self.max = max
        self.name = name

    def solution_value(self):
        return self.max


class MockConstraint(DTO):
    def __init__(self, min: float, max: float, coeffs=None):
        self.min = min
        self.max = max
        self.coeffs = coeffs if coeffs else {}

    def SetCoefficient(self, var: MockNumVar, cost: int):
        self.coeffs[var] = cost


class MockObjective(DTO):
    def __init__(self, min=False, coeffs=None):
        self.min = min
        self.coeffs = coeffs if coeffs else {}

    def SetMinimization(self):
        self.min = True

    def SetCoefficient(self, var: MockNumVar, cost: int):
        self.coeffs[var] = cost

    def Value(self):
        return 0


class MockSolver:
    def __init__(self):
        pass

    def NumVar(self, min: float, max: float, name: str = ''):
        return MockNumVar(min, max, name)

    def Objective(self) -> MockObjective:
        return MockObjective()

    def Constraint(self, min: int, max: int):
        return MockConstraint(min, max)

    def Solve(self):
        pass

    def EnableOutput(self):
        pass

    def ExportModelAsLpFormat(self, toggle: bool):
        return ''



import uuid
from copy import deepcopy
from typing import List, Tuple

import pandas as pd
import numpy as np
from hadar.solver.actor.domain.message import Exchange


class Ledger:
    """Meta Ledger to implement eq, hash, str, repr method"""
    def __init__(self, headers: List[List[str]]):
        self.ledger = pd.DataFrame(columns=[name for name, dtype in headers])  #NOSONAR
        self.ledger = self.ledger.astype({name: dtype for name, dtype in headers})

    def __hash__(self):
        return hash(self.ledger)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.ledger.equals(other.ledger)

    def __str__(self):
        return str(self.ledger)

    def __repr__(self):
        return self.__str__()


class LedgerExchange(Ledger):
    """
    Manage exchange ledger for a dispatcher.
    """
    TYPES = ['import', 'export', 'transfer']

    def __init__(self):
        Ledger.__init__(self, [['prod_type', 'object'], ['border', 'object'],
                               ['quantity', 'int64'], ['path_node', 'object'], ['type', 'object']])

    def add_all(self, ex: List[Exchange], type: str):
        """
        Add many exchanges to ledger

        :param ex: list of exchange to add
        :param type: exchange type [import/export/transfer]
        """
        for e in ex:
            self.add(e, type)

    def add(self, ex: Exchange, type: str):
        """
        Add exchange to ledger

        :param ex: exchange object to add
        :param type: exchange type [import/export/transfer]
        :return:
        """
        if type not in LedgerExchange.TYPES:
            raise ValueError("exchange type should be [import/export/transfer], you give {}".format(type))
        if ex.id in self.ledger.index:
            raise ValueError('Exchange already stored in ledger')
        border = ex.path_node[0]
        self.ledger.loc[ex.id] = [ex.production_type,  border, ex.quantity, ex.path_node, type]

    def delete(self, ex: Exchange):
        """
        Delete exchange from ledger.

        :param ex: exchange object to delete
        :return:
        """
        self.ledger.drop(ex.id, inplace=True)

    def delete_all(self, exs: List[Exchange]):
        """
        Delete many exchanges from ledger.

        :param exs: exchanges list to delete
        :return:
        """
        ids = [ex.id for ex in exs]
        self.ledger.drop(ids, inplace=True)

    def get_exchanges(self, ids: List[int]) -> List[Exchange]:
        """
        Get exchanges by id
        :param ids: exchange ids
        :return: slice of ledger with asked exchanges
        """
        return [Exchange(id=id, quantity=qt, production_type=prod_type, path_node=path)
                for id, (prod_type, border, qt, path, _) in self.ledger.loc[ids].iterrows()]

    def sum_production(self, production_type):
        """
        Sum production quantity used.

        :param production_type: production type
        :return: quantity produce by this production
        """
        return self.ledger[(self.ledger['prod_type'] == production_type) & (self.ledger['type'] == 'export')]['quantity'].sum()

    def sum_border(self, name: str):
        """
        Sum all quantity send to a border.

        :param name: border's name
        :return: quantity send to this border
        """
        return self.ledger[self.ledger['border'] == name]['quantity'].sum()

    def filter_production_available(self, productions: pd.DataFrame) -> pd.DataFrame:
        """
        Filter production DataFrame with currently production in exchange

        :param productions: production DataFrame like production ledger
        :return: production DataFrame without production present in ledger exchange
        """
        prod_not_available = productions.index.intersection(self.ledger.index)
        return productions.drop(prod_not_available)

class LedgerProduction(Ledger):
    """Manage production used by dispatcher"""

    def __init__(self, uuid_generate=uuid.uuid4):
        self.uuid_generate = uuid_generate
        Ledger.__init__(self, [['cost', 'int64'], ['quantity', 'int64'], ['type', 'object'],
                               ['used', 'bool'], ['path_node', 'object']])

    def __eq__(self, other):
        if not hasattr(other, 'ledger'):
            return False
        return self.ledger.equals(other.ledger)

    def add_production(self, cost: int, quantity: int, type: str = '', used: bool = False):
        """
        Add production from internal.

        :param cost: production cost
        :param quantity: production quantity
        :param type: production type
        :param used: set used or not
        :return:
        """
        append = pd.DataFrame(data={
            'cost': np.ones(quantity, dtype=int) * cost,
            'quantity': np.ones(quantity, dtype=int),
            'type': [type] * quantity,
            'used': [used] * quantity,
            'path_node': [None] * quantity},
            index=[self.uuid_generate() for _ in range(quantity)])

        self.ledger = pd.concat([self.ledger, append])


    def add_exchanges(self, cost: int, ex: List[Exchange], used: bool = False):
        """
        Add production from external.

        :param cost: cost of external production
        :param used: set used or not
        :param ex: exchange object where production comes
        :return:
        """
        quantity = len(ex)
        append = pd.DataFrame(data={
            'cost': np.ones(quantity, dtype=int) * cost,
            'quantity': [e.quantity for e in ex],
            'type': ['import'] * quantity,
            'used': [used] * quantity,
            'path_node': [e.path_node for e in ex]},
            index=[e.id for e in ex])

        self.ledger = pd.concat([self.ledger, append])

    def delete(self, id: uuid):
        """
        Delete production by its id.

        :param id: production id
        :return:
        """
        self.ledger.drop(id, inplace=True)

    def delete_all(self, ids):
        """
        Delete many production by their ids.

        :param ids: productions id
        :return:
        """
        self.ledger.drop(ids, inplace=True)

    def filter_exchanges(self) -> pd.DataFrame:
        """
        Get only external production.

        :return: dataframe with external production
        """
        return self.ledger[self.ledger['path_node'].notnull()]

    def filter_useless_exchanges(self) -> pd.DataFrame:
        """
        Get production from exchange tagged as not used.

        :return: DataFrame with only exchange not used
        """
        return self.ledger[self.ledger['path_node'].notnull() & ~self.ledger['used']]

    def group_free_productions(self) -> pd.DataFrame:
        """
        Regroup free productions by type.
        :return: dataframe with each production type sum up on one row (quantities are sum up)
        """
        df = self.ledger[self.ledger['path_node'].isnull() & ~self.ledger['used']]
        return pd.pivot_table(df, values=['cost', 'quantity'], index=['type'], aggfunc={'cost': np.mean, 'quantity': np.sum})

    def get_free_productions_by_type(self, type: str) -> pd.DataFrame:
        """
        Get a slice of ledger with internal productions with same type.

        :param type: type of production asked
        :return: slice of global ledger
        """
        slice = self.ledger[(self.ledger['path_node'].isnull()) & (~self.ledger['used']) & (self.ledger['type'] == type)]
        return deepcopy(slice)

    def get_production_quantity(self, type: str, used: bool) -> int:
        """
        Get total quantity production if used or not.

        :param type: type of production
        :param used: get used or free quantity
        :return:
        """
        return self.ledger[(self.ledger['type'] == type) & (self.ledger['used'] == used)]['quantity'].sum()


    def apply_adequacy(self, quantity: int):
        """
        Split current production state between used and free production according quantity consumption needed.

        :param quantity: consumption quantity
        :return: new cost, power used
        """
        used, free = LedgerProduction.split_by_quantity(self.ledger, quantity)
        used['used'] = True
        free['used'] = False
        self.ledger = pd.concat([used, free])

        cost = (used['cost'] * used['quantity']).sum()
        power = used['quantity'].sum()
        return cost, power


    @classmethod
    def split_by_quantity(cls, prod: pd.DataFrame, quantity: int) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Split DataFrame between production used and productions free according quantity production asked.

        :return: DataFrame productions used, DataFrame productions free
        """
        prod = prod.sort_values(by=['cost', 'quantity'], ascending=[True, False])
        index_split = prod['quantity'].cumsum() <= quantity
        return prod[index_split], prod[~index_split]



class LedgerConsumption(Ledger):
    """Manage consumption used by dispatcher"""

    def __init__(self):
        Ledger.__init__(self, [['cost', 'int64'], ['quantity', 'int64']])

    def add(self, type: str, cost: int, quantity: int):
        self.ledger.loc[type] = [cost, quantity]
        self.ledger.sort_values(by='cost', ascending=False, inplace=True)


    def delete(self, type: str):
        self.ledger.drop(type, inplace=True)

    def find_consumption(self, type: str) -> pd.Series:
        return self.ledger.loc[type]

    def sum_quantity(self):
        return self.ledger['quantity'].sum()

    def compute_cost(self, quantity: int) -> int:
        if self.ledger.size == 0:
            return 0
        cum = self.ledger['quantity'].cumsum()
        rac = cum - quantity
        rac[rac < 0] = 0
        unit = rac.diff()
        unit.iloc[0] = rac.iloc[0]

        return (self.ledger['cost']*unit).sum()


class LedgerBorder(Ledger):
    """Manage borders used by dispatcher"""

    def __init__(self):
        Ledger.__init__(self, [['cost', 'int64'], ['quantity', 'int64']])

    def __eq__(self, other):
        if not hasattr(other, 'ledger'):
            return False
        return self.ledger.equals(other.ledger)

    def add(self, dest: str, cost: int, quantity: int):
        self.ledger.loc[dest] = [cost, quantity]

    def delete(self, dest: str):
        self.ledger.drop(dest, inplace=True)

    def find_border(self, dest: str):
        return self.ledger.loc[dest]

    def find_border_in_path(self, path: List[str]):
        return self.ledger[self.ledger.index.isin(path)].iloc[0]



from copy import deepcopy
from typing import Any, Tuple

from hadar.solver.actor.handler.handler import State
from hadar.solver.actor.handler.handler import *


class StartHandler(Handler):
    """
    When Start message receive:
    --> propose free production
    """

    def __init__(self, params: HandlerParameter):
        """
        Initiate handler.

        :param params: handler parameters to use
        """
        Handler.__init__(self, params=params)
        self.handler = ProposeFreeProductionHandler(params=params, next=ReturnHandler())

    def execute(self, state: State, message=None) -> Tuple[State, Any]:
        """
        Execute handler.

        :param state: current actor state
        :param message: Start message
        :return: next actor state
        """
        return self.handler.execute(deepcopy(state), deepcopy(message))


class CanceledCustomerExchangeHandler(Handler):
    """
    When CancelCustomerExchangeHandler
    --> cancel exportation
    -----> if it's node producer : propose free production
    """
    def __init__(self, params: HandlerParameter):
        """
        Initiate handler.

        :param params: handler parameters to use
        """
        Handler.__init__(self, params=params)
        self.handler = CancelExportationHandler(params=params,
                    next=BackwardMessageHandler(type='tell',
                            after_backward=ReturnHandler(),
                            on_resume=ProposeFreeProductionHandler(next=ReturnHandler()),
                    ))

    def execute(self, state: State, message=None) -> Tuple[State, Any]:
        """
        Execute handler.

        :param state: current actor state
        :param message: CancelCustomerExchange message
        :return: next actor state
        """
        return self.handler.execute(deepcopy(state), deepcopy(message))


class ProposalOfferHandler(Handler):
    """
    When receive proposal offer:
    --> check border capacity
    ----> if node is transfer, wait response
    ------> save final exchanges
    ----> if node is producer, check production capacity
    ------> save final exchanges
    """
    def __init__(self, params: HandlerParameter):
        """
        Initiate handler.

        :param params: handler parameters to use
        """
        Handler.__init__(self, params=params)
        self.handler = CheckOfferBorderCapacityHandler(params=params,
                    next=BackwardMessageHandler(type='ask',
                        after_backward=SaveExchangeHandler(exchange_type='transfer', next=ReturnHandler()),
                        on_resume=AcceptExchangeHandler(
                            next=SaveExchangeHandler(exchange_type='export', next=ReturnHandler()))
                    ))

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Execute handler.

        :param state: current actor state
        :param message: ProposalOffer message
        :return: next actor state, list of exchanges
        """
        return self.handler.execute(deepcopy(state), deepcopy(message))


class ProposalHandler(Handler):
    """
    When receive proposal:
    -> compare proposal with current state
    ---> if useless
    -----> forward proposal
    ---> if useful
    -----> make an offer, wait response
    -------> save exchanges accepted by producer
    ---------> compute new adequacy
    -----------> propose free production
    -------------> cancel useless importation
    """
    def __init__(self, params: HandlerParameter):
        """
        Initiate handler.

        :param params: handler parameters to use
        """
        Handler.__init__(self, params=params)
        self.handler = CompareNewProduction(params=params,
                            for_prod_useless=ForwardMessageHandler(next=ReturnHandler()),
                            for_prod_useful=MakerOfferHandler(
                                next=SaveExchangeHandler(exchange_type='import',
                                    next=AdequacyHandler(
                                        next=ProposeFreeProductionHandler(
                                            next=CancelUselessImportationHandler(next=ReturnHandler())
                                        )
                                    )
                                )
                            ))

    def execute(self, state: State, message: Any = None) -> Tuple[State, Any]:
        """
        Execute handler.

        :param state: current actor state
        :param message: ProposalOffer message
        :return: next actor state
        """
        return self.handler.execute(deepcopy(state), deepcopy(message))



from ortools.linear_solver.pywraplp import Solver

from hadar.solver.input import InputNode, Study
from hadar.solver.lp.domain import *
from hadar.solver.output import OutputNode, Result


class InputMapper:
    """
    Input mapper from global domain to linear programming specific domain
    """

    def __init__(self, solver: Solver, study: Study):
        """
        Instantiate mapper.

        :param solver: ortools solver to used to create variables
        :param study: study data
        """
        self.solver = solver
        self.study = study

    def get_var(self, name: str, t: int) -> LPNode:
        """
        Map InputNode to LPNode.

        :param name: node name
        :param t: timestamp
        :return: LPNode according to node name at t in study
        """
        consumptions = [LPConsumption(type=c.type, cost=float(c.cost), quantity=c.quantity[t],
                                      variable=self.solver.NumVar(0, float(c.quantity[t]), name='lol {} on {} at {}'.format(c.type, name, t)))
                        for c in self.study.nodes[name].consumptions]

        productions = [LPProduction(type=p.type, cost=float(p.cost), quantity=p.quantity[t],
                                    variable=self.solver.NumVar(0, float(p.quantity[t]), 'prod {} on {} at {}'.format(p.type, name, t)))
                       for p in self.study.nodes[name].productions]

        borders = [LPBorder(dest=b.dest, cost=float(b.cost), src=name, quantity=b.quantity[t],
                            variable=self.solver.NumVar(0, float(b.quantity[t]), 'border on {} to {} at {}'.format(name, b.dest, t)))
                   for b in self.study.nodes[name].borders]

        return LPNode(consumptions=consumptions, productions=productions, borders=borders)


class OutputMapper:
    """
    Output mapper from specific linear programming domain to global domain.
    """
    def __init__(self, solver: Solver, study: Study):
        """
        Instantiate mapper.

        :param solver: ortools solver to use to fetch variable value
        :param study: input study to reproduce structure
        """
        self.nodes = {name: OutputNode.build_like_input(input) for name, input in study.nodes.items()}

    def set_var(self, name: str, t: int, vars: LPNode):
        """
        Map linear programming node to global node (set inside intern attribute).

        :param name: node name
        :param t: timestamp index
        :param vars: linear programming node with ortools variables inside
        :return: None (use get_result)
        """
        for i in range(len(vars.consumptions)):
            self.nodes[name].consumptions[i].quantity[t] = vars.consumptions[i].quantity - vars.consumptions[i].variable.solution_value()

        for i in range(len(vars.productions)):
            self.nodes[name].productions[i].quantity[t] = vars.productions[i].variable.solution_value()

        for i in range(len(vars.borders)):
            self.nodes[name].borders[i].quantity[t] = vars.borders[i].variable.solution_value()

    def get_result(self) -> Result:
        """
        Get result.

        :return: final result after map all nodes
        """
        return Result(nodes=self.nodes)



import numpy as np
from typing import Union, List

from ortools.linear_solver.pywraplp import Variable

from hadar.solver.input import DTO


class LPConsumption(DTO):
    """
    Consumption element for linear programming.
    """

    def __init__(self, quantity: int, variable: Variable, cost: float = 0, type: str = ''):
        """
        Instance consumption.

        :param quantity: quantity to match
        :param variable: ortools variables. represent a loss of load inside equation
        :param cost: unavailability cost
        :param type: consumption type name
        """
        self.cost = cost
        self.quantity = quantity
        self.type = type
        self.variable = variable


class LPProduction(DTO):
    """
    Production element for linear programming.
    """

    def __init__(self, quantity: int, variable: Variable, cost: float = 0, type: str = 'in'):
        """
        Instance production.

        :param quantity: production capacity
        :param variable: ortools variables. Represent production used inside equation
        :param cost: cost of use
        :param type: production type name
        """
        self.type = type
        self.cost = cost
        self.variable = variable
        self.quantity = quantity


class LPBorder(DTO):
    """
    Border element for linear programming
    """
    def __init__(self, src: str, dest: str, quantity: int, variable: Variable, cost: float = 0):
        """
        Instance border.

        :param src: node source name
        :param dest: node destination name
        :param quantity: border capacity
        :param variable: ortools variables. Represent border used inside equation
        :param cost: cost of use
        """
        self.src = src
        self.dest = dest
        self.quantity = quantity
        self.variable = variable
        self.cost = cost


class LPNode(DTO):
    """
    Node element for linear programming
    """
    def __init__(self, consumptions: List[LPConsumption], productions: List[LPProduction], borders: List[LPBorder]):
        """
        Instance node.

        :param consumptions: consumptions list
        :param productions: productions list
        :param borders: border list
        """
        self.consumptions = consumptions
        self.productions = productions
        self.borders = borders


import unittest

from ortools.linear_solver import pywraplp
from hadar.solver.lp.domain import *
from hadar.solver.lp.mapper import InputMapper, OutputMapper
from hadar.solver.input import *
from hadar.solver.output import *
from tests.solver.lp.ortools_mock import MockSolver, MockNumVar
from tests.utils import assert_study


class TestInputMapper(unittest.TestCase):
    def test_map_input(self):
        # Input
        study = Study(['a', 'be']) \
            .add_on_node('a', Consumption(type='load', quantity=[10, 1], cost=10)) \
            .add_on_node('a', Production(type='nuclear', quantity=[12, 2], cost=10)) \
            .add_border(src='a', dest='be', quantity=[10, 3], cost=2)

        s = MockSolver()

        mapper = InputMapper(solver=s, study=study)

        # Expected
        out_cons_0 = [LPConsumption(type='load', cost=10, quantity=10, variable=MockNumVar(0, 10, 'lol load on a at 0'))]
        out_prod_0 = [LPProduction(type='nuclear', cost=10, quantity=12, variable=MockNumVar(0, 12.0, 'prod nuclear on a at 0'))]

        out_bord_0 = [LPBorder(src='a', dest='be', cost=2, quantity=10, variable=MockNumVar(0, 10.0, 'border on a to be at 0'))]
        out_node_0 = LPNode(consumptions=out_cons_0, productions=out_prod_0, borders=out_bord_0)

        self.assertEqual(out_node_0, mapper.get_var(name='a', t=0))

        out_cons_1 = [LPConsumption(type='load', cost=10, quantity=1, variable=MockNumVar(0, 1, 'lol load on a at 1'))]
        out_prod_1 = [LPProduction(type='nuclear', cost=10, quantity=2, variable=MockNumVar(0, 2.0, 'prod nuclear on a at 1'))]

        out_bord_1 = [LPBorder(src='a', dest='be', cost=2, quantity=3, variable=MockNumVar(0, 3.0, 'border on a to be at 1'))]
        out_node_1 = LPNode(consumptions=out_cons_1, productions=out_prod_1, borders=out_bord_1)

        self.assertEqual(out_node_1, mapper.get_var(name='a', t=1))


class TestOutputMapper(unittest.TestCase):
    def test_map_output(self):
        # Input
        study = Study(['a', 'be']) \
            .add_on_node('a', Consumption(type='load', quantity=[10, 20], cost=10)) \
            .add_on_node('a', Production(type='nuclear', quantity=[12, 2], cost=10)) \
            .add_border(src='a', dest='be', quantity=[10, 3], cost=2)

        s = MockSolver()
        mapper = OutputMapper(solver=s, study=study)

        out_cons_0 = [LPConsumption(type='load', cost=10, quantity=10, variable=MockNumVar(0, 5.0, 'lol load on a'))]
        out_prod_0 = [LPProduction(type='nuclear', cost=10, quantity=12, variable=MockNumVar(0, 12.0, 'prod nuclear on a'))]

        out_bord_0 = [LPBorder(src='a', dest='be', cost=2, quantity=10, variable=MockNumVar(0, 8.0, 'border on a to be'))]
        mapper.set_var(name='a', t=0, vars=LPNode(consumptions=out_cons_0 , productions=out_prod_0, borders=out_bord_0))

        out_cons_1 = [LPConsumption(type='load', cost=10, quantity=20, variable=MockNumVar(0, 5.0, 'lol load on a'))]
        out_prod_1 = [LPProduction(type='nuclear', cost=10, quantity=2, variable=MockNumVar(0, 112.0, 'prod nuclear on a'))]

        out_bord_1 = [LPBorder(src='a', dest='be', cost=2, quantity=10, variable=MockNumVar(0, 18.0, 'border on a to be'))]
        mapper.set_var(name='a', t=1, vars=LPNode(consumptions=out_cons_1 , productions=out_prod_1, borders=out_bord_1))

        # Expected
        node = OutputNode(consumptions=[OutputConsumption(type='load', quantity=[5, 15], cost=10)],
                          productions=[OutputProduction(type='nuclear', quantity=[12, 112], cost=10)],
                          borders=[OutputBorder(dest='be', quantity=[8, 18], cost=2)])
        expected = Result(nodes={'a': node, 'be': OutputNode(consumptions=[], productions=[], borders=[])})


        assert_study(self, expected=expected, result=mapper.get_result())



from hadar.solver.input import Study
from hadar.solver.actor.solver import solve as solve_actor
from hadar.solver.output import Result
from hadar.solver.lp.solver import solve_lp


def solve(study: Study, kind: str = 'lp') -> Result:
    """
    Solve adequacy study.

    :param study: study to resolve
    :param kind: type of solver to use. 'actor' or 'lp'
    :return: study's result
    """
    if kind == 'actor':
        return solve_actor(study)
    if kind == 'lp':
        return solve_lp(study)
    raise ValueError('kind {} not supported. Support only [actor]')



import unittest
from unittest.mock import MagicMock

from hadar.solver.lp.solver import *
from hadar.solver.input import *
from hadar.solver.output import *
from hadar.solver.lp.solver import _solve
from tests.solver.lp.ortools_mock import *


class TestObjectiveBuilder(unittest.TestCase):

    def test_add_node(self):
        # Mock
        objective = MockObjective()
        solver = MockSolver()

        # Input
        consumptions = [LPConsumption(type='load', quantity=10, cost=10, variable=MockNumVar(0, 10, 'load'))]
        productions = [LPProduction(type='solar', quantity=10, cost=20, variable=MockNumVar(0, 20, 'solar'))]
        borders = [LPBorder(src='fr', dest='be', quantity=10, cost=30, variable=MockNumVar(0, 30, 'be'))]
        node = LPNode(consumptions=consumptions, productions=productions, borders=borders)

        # Expected
        coeffs = {MockNumVar(0, 10, 'load'): 10, MockNumVar(0, 20, 'solar'): 20, MockNumVar(0, 30, 'be'): 30}
        expected = MockObjective(min=True, coeffs=coeffs)

        # Test
        builder = ObjectiveBuilder(solver=solver)
        builder.add_node(node)
        builder.build()

        self.assertEqual(expected, builder.objective)


class TestAdequacyBuilder(unittest.TestCase):

    def test_add_node(self):
        # Mock
        solver = MockSolver()

        # Input
        fr_consumptions = [LPConsumption(type='load', quantity=10, cost=10, variable=MockNumVar(0, 10, 'load'))]
        fr_productions = [LPProduction(type='solar', quantity=10, cost=20, variable=MockNumVar(0, 20, 'solar'))]
        fr_borders = [LPBorder(src='fr', dest='be', quantity=10, cost=30, variable=MockNumVar(0, 30, 'be'))]
        fr_node = LPNode(consumptions=fr_consumptions, productions=fr_productions, borders=fr_borders)

        be_node = LPNode(consumptions=[], productions=[], borders=[])

        # Expected
        fr_coeffs = {MockNumVar(0, 10, 'load'): 1, MockNumVar(0, 20, 'solar'): 1, MockNumVar(0, 30, 'be'): -1}
        fr_constraint = MockConstraint(10, 10, coeffs=fr_coeffs)

        be_coeffs = {MockNumVar(0, 30, 'be'): 1}
        be_constraint = MockConstraint(0, 0, coeffs=be_coeffs)

        # Test
        builder = AdequacyBuilder(solver=solver, horizon=1)
        builder.add_node(name='fr', node=fr_node, t=0)
        builder.add_node(name='be', node=be_node, t=0)
        builder.build()

        self.assertEqual(fr_constraint, builder.constraints[0]['fr'])
        self.assertEqual(be_constraint, builder.constraints[0]['be'])


class TestSolve(unittest.TestCase):
    def test_solve(self):
        # Input
        study = Study(node_names=['a']) \
            .add_on_node(node='a', data=Consumption(type='load', cost=10, quantity=[10]))

        # Expected
        out_a = OutputNode(consumptions=[OutputConsumption(type='load', cost=10, quantity=[10])],
                       productions=[], borders=[])
        exp_result = Result(nodes={'a': out_a})

        # Mock
        solver = MockSolver()
        solver.Solve = MagicMock()

        objective = ObjectiveBuilder(solver=solver)
        objective.add_node = MagicMock()
        objective.build = MagicMock()

        adequacy = AdequacyBuilder(solver=solver, horizon=study.horizon)
        adequacy.add_node = MagicMock()
        adequacy.build = MagicMock()

        in_cons = LPConsumption(type='load', quantity=10, cost=10, variable=MockNumVar(0, 10, 'load'))
        var = LPNode(consumptions=[in_cons], productions=[], borders=[])
        in_mapper = InputMapper(solver=solver, study=study)
        in_mapper.get_var = MagicMock(return_value=var)

        out_mapper = OutputMapper(solver=solver, study=study)
        out_mapper.set_var = MagicMock()
        out_mapper.get_result = MagicMock(return_value=exp_result)


        # Test
        res = _solve(study, solver, objective, adequacy, in_mapper, out_mapper)

        self.assertEqual(exp_result, res)

        in_mapper.get_var.assert_called_with(name='a', t=0)
        adequacy.add_node.assert_called_with(name='a', t=0, node=var)
        objective.add_node.assert_called_with(node=var)

        objective.build.assert_called_with()
        adequacy.build.assert_called_with()

        solver.Solve.assert_called_with()

        out_mapper.set_var.assert_called_with(name='a', t=0, vars=var)



import setuptools
import hadar

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    dependencies = f.read().split('\n')

setuptools.setup(
    name="hadar",
    version=hadar.__version__,
    author="RTE France",
    author_email="francois.jolain@rte-international.com",
    description="python adequacy library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hadar-simulator/hadar",
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import uuid
from typing import List


class Message:
    """
    Implement basic method for messages send to actor
    """
    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.viewitems())))

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.__dict__ == other.__dict__

    def __str__(self):
        return "{}({})".format(type(self).__name__, ", ".join(["{}={}".format(k, str(self.__dict__[k])) for k in sorted(self.__dict__)]))

    def __repr__(self):
        return self.__str__()


class Exchange(Message):
    """Exchange token to be send"""
    def __init__(self, quantity=0, id: uuid = 0, production_type: str = '', path_node: List[str] = []):
        """
        Create instance exchange message.

        :param quantity: quality of items to exchanges. Quantity is not divisible.
        :param id: unique id of this exchanges. It's the same as production token
        :param production_type: type of production used
        :param path_node: path node inside network te retrieve producer
        """
        self.quantity = quantity
        self.id = id
        self.production_type = production_type
        self.path_node = path_node


class Proposal(Message):
    """Proposal capacity production to be spread accros network"""
    def __init__(self, production_type: str, cost: int, quantity: int, path_node: List[str]):
        """
        Create instance.

        :param production_type: production type available
        :param cost: cost of production
        :param quantity: sum of quantity available
        :param path_node: path node inside network to retrieve producer
        """
        self.production_type = production_type
        self.cost = cost
        self.quantity = quantity
        self.path_node = path_node


class ProposalOffer(Proposal):
    """Offer send by consumer as a response to a proposal from a producer"""
    def __init__(self, production_type: str, cost: int, quantity: int, path_node: List[str], return_path_node: List[str]):
        """
        Create instance.

        :param production_type: production type asked
        :param cost: cost of productions asked (added with border cost during transfer)
        :param quantity: quantity asked
        :param path_node: path node inside network to retrieve producer
        :param return_path_node: path node inside network to retrieve consumer
        """
        Proposal.__init__(self, production_type, cost, quantity, path_node)
        self.return_path_node = return_path_node


class ConsumerCanceledExchange(Message):
    """Order send by consumer to cancel exchange"""
    def __init__(self, exchanges: List[Exchange], path_node: List[str]=[]):
        """
        Create instance.

        :param exchanges: exchanges to cancel
        :param path_node: path node to follow to retrieve producer
        """
        self.exchanges = exchanges
        self.path_node = path_node


class Snapshot(Message):
    """Message send by orchestrator to backup a snapshot of an actor state"""
    def __init__(self):
        pass  # Standalone message


class Start(Message):
    """Message send by orchestrator to start exchanges between actors"""
    def __init__(self):
        pass  # Standalone message


class Next(Message):
    """Message send by orchestrator to actor to force to go to the next timestamp"""
    def __init__(self):
        pass  # Standalone message



import logging
import os
import sys

__version__ = '0.1.2'

level = os.getenv('HADAR_LOG', 'WARNING')

if level == 'INFO':
    level = logging.INFO
elif level == 'DEBUG':
    level = logging.DEBUG
elif level == 'WARNING':
    level = logging.WARNING
elif level == 'ERROR':
    level = logging.ERROR
else:
    level = logging.WARNING

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(formatter)
logging.basicConfig(level=level, handlers=[handler])
